"""
  Deterministic Optimistic Optimization (DOO) and its parallel version (PDOO)
  Adapted from MFTREE_DET library -- https://github.com/rajatsen91/MFTREE_DET
  References:
    - Sen et al, 2018. Multi-fidelity Black-box Optimization with Hierarchical Partitions
      (http://proceedings.mlr.press/v80/sen18a.html)
    - Grill et al, 2015: Black-box optimization of noisy functions with Unknown
      Smoothness (https://dl.acm.org/citation.cfm?id=2969314)
  -- kvysyara@andrew.cmu.edu
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name
# pylint: disable=no-self-use

from __future__ import division

from functools import total_ordering
from argparse import Namespace
try:
  import Queue as Qu
except ImportError:
  import queue as Qu
import numpy as np

from .general_utils import map_to_bounds


class DOOFunction(object):
  """ This just creates a wrapper to call the function by appropriately creating bounds
      and querying appropriately. """

  def __init__(self, func, domain_bounds, vectorised=False):
    """ Constructor.
        func: takes an argument x which is a point in the domain.
        domain_bounds: are the bounds of the domain space
        vectorised: If True it means func can take multiple inputs and produce multiple
                    outputs. If False, the functions can take only single inputs in
                    'column' form.
    """
    self.func = func
    self.domain_bounds = np.array(domain_bounds)
    self.domain_dim = len(domain_bounds)
    self.vectorised = vectorised

  def get_unnormalised_coords(self, X):
    """ Maps points in the cube to the original space. """
    ret_X = None if X is None else map_to_bounds(X, self.domain_bounds)
    return ret_X

  def eval_single_point(self, X):
    """ Evaluates X at the given Z at a single point. """
    if not self.vectorised:
      return float(self.func(X))
    else:
      X = np.array(X).reshape((1, self.domain_dim))
      return float(self.func(X))

  def eval_single_point_normalised(self, X):
    """ Evaluates X at the given Z at a single point using normalised coordinates. """
    X = self.get_unnormalised_coords(X)
    return self.eval_single_point(X)

  def eval_cost_single_point_normalised(self, Z):
    """ Evaluates the cost function at a single point using normalised coordinates. """
    # pylint: disable=unused-argument
    return 1.0


@total_ordering
class Node(object):
  """ Base object for the node of a tree. """

  def __init__(self, cell, value, fidel, upp_bound, height, dimension):
    """ cell: tuple """
    self.cell = cell
    self.value = value
    self.fidelity = fidel
    self.upp_bound = upp_bound
    self.height = height
    self.dimension = dimension

  def __cmp__(self, other):
    return cmp(other.upp_bound, self.upp_bound)

  def __lt__(self, other):
    return other.upp_bound < self.upp_bound

  def __eq__(self, other):
    return other.upp_bound == self.upp_bound


class OptTree(object):
  """ Class for Tree search based optimisation. """

  def __init__(self, doo_obj, nu_max, rho_max, total_budget, K, C_init, tol=1e-3,
               Randomize=False):
    self.doo_obj = doo_obj
    self.nu_max = nu_max
    self.rho_max = rho_max
    self.total_budget = total_budget
    self.K = K
    self.C = C_init
    self.tol = tol
    self.value_dict = {}
    self.qcount = 0
    self.last_z_hist = []
    self.Randomize = Randomize
    self.query_vals = []
    self.query_pts = []

  def get_value(self, cell, fidel):
    """ cell: tuple """
    # pylint: disable=unused-argument
    self.qcount = self.qcount + 1
    x = np.array([(s[0]+s[1])/2.0 for s in list(cell)])
    y = self.doo_obj.eval_single_point_normalised(x)
    if len(self.query_vals) <= self.total_budget:
      self.query_pts.append(x)
      self.query_vals.append(y)
    return y

  def get_queried_pts(self):
    """ Returns queried points and corresponding values. """
    return self.query_pts, self.query_vals

  def querie(self, cell, height, rho, nu, dimension, option=1):
    """ Query. """
    diam = nu*(rho**height)
    if option == 1:
      z = min(max(1 - diam/self.C, self.tol), 1.0)
    else:
      z = 1.0
    if cell in self.value_dict:
      current = self.value_dict[cell]
      if abs(current.fidelity - z) <= self.tol:
        value = current.value
        cost = 0
      else:
        value = self.get_value(cell, z)
        self.last_z_hist = self.last_z_hist + [z]
        if abs(value - current.value) > self.C*abs(current.fidelity - z):
          self.C = 2.0*self.C
        current.value = value
        current.fidelity = z
        self.value_dict[cell] = current
        cost = self.doo_obj.eval_cost_single_point_normalised([z])
    else:
      value = self.get_value(cell, z)
      self.last_z_hist = self.last_z_hist + [z]
      bhi = diam + self.C*(1.0 - z) + value
      self.value_dict[cell] = Node(cell, value, z, bhi, height, dimension)
      cost = self.doo_obj.eval_cost_single_point_normalised([z])

    bhi = diam + self.C*(1.0 - z) + value
    current_object = Node(cell, value, z, bhi, height, dimension)
    return current_object, cost

  def split_children(self, current, rho, nu, option=1):
    """ Split children. """
    pcell = list(current.cell)
    span = [abs(pcell[i][1] - pcell[i][0]) for i in range(len(pcell))]
    if self.Randomize:
      dimension = np.random.choice(range(len(pcell)))
    else:
      dimension = np.argmax(span)
    dd = len(pcell)
    if dimension == current.dimension:
      dimension = (current.dimension - 1)%dd
    cost = 0
    h = current.height + 1
    l = np.linspace(pcell[dimension][0], pcell[dimension][1], self.K+1)
    children = []
    for i in range(len(l)-1):
      cell = []
      for j, _ in enumerate(pcell):
        if j != dimension:
          cell = cell + [pcell[j]]
        else:
          cell = cell + [(l[i], l[i+1])]
      child, c = self.querie(tuple(cell), h, rho, nu, dimension, option)
      children = children + [child]
      cost = cost + c

    return children, cost

  def run_DOO(self, budget, nu, rho):
    """ Runs DOO optimisation. """
    leaf_Q = Qu.PriorityQueue()
    d = self.doo_obj.domain_dim
    cell = tuple([tuple([0, 1]) for _ in range(d)])
    height = 0
    cost = 0
    current, c = self.querie(cell, height, rho, nu, 0)
    cost = cost + c
    leaf_Q.put(current)
    dict_of_points = {}
    while cost <= budget:
      current = leaf_Q.get()
      dict_of_points[current.cell] = {'val':current.value, 'fidel': current.fidelity,
                                      'height':current.height}
      children, curr_cost = self.split_children(current, rho, nu)
      if current.cell == children[0].cell:
        break
      cost = cost + curr_cost
      for child in children:
        leaf_Q.put(child)
    while not leaf_Q.empty():
      c = leaf_Q.get()
      dict_of_points[c.cell] = {'val':c.value, 'fidel': c.fidelity, 'height':c.height}

    #maxi = float(-sys.maxint - 1)
    maxi = float('-inf')
    point = 0
    maxh = 0
    val = 0
    fidel = 0
    for key in dict_of_points:
      c = dict_of_points[key]
      if c['val'] - self.C*(1.0 - c['fidel']) > maxi:    #- nu*(rho**c.height) > maxi:
        maxi = c['val'] - self.C*(1.0 - c['fidel'])      #- nu*(rho**c.height)
        val = c['val']
        fidel = c['fidel']
        point = np.array([(s[0]+s[1])/2 for s in key])
        maxh = c['height']

    return val, fidel, point, cost, maxh

  def run_PDOO(self, mult=0.5):
    """ Runs DOO optimisation iteratively based on the budget. """
    Dm = int(np.log(self.K)/np.log(1/self.rho_max))
    n = self.total_budget / \
        self.doo_obj.eval_cost_single_point_normalised(np.array([1.0]))
    N = int(mult * Dm * np.log(n/np.log(n)))
    budget = self.total_budget/float(N)
    nu = self.nu_max
    total_cost = 0.0
    results = []
    for i in range(N):
      rho = (self.rho_max)**(float(N)/(N-i))
      est, fidel, point, cost, h = self.run_DOO(budget, nu, rho)
      results = results + [(est, fidel, point, cost, h)]
      total_cost = total_cost + cost
    temp = [s[0] - self.C*(1 - s[1]) for s in results]
    index = np.argmax(temp)
    self.last_results = (temp[index], results[index][2], total_cost)

    return results, index

def pdoo_wrap(doo_obj, total_budget, nu_max=1.0, rho_max=0.9, K=2, C_init=0.8, tol=1e-3,
              POO_mult=0.5, Randomize=False, return_history=False):
  """ Wrapper for running PDOO optimisation. """
  # pylint: disable=too-many-locals
  total_budget = total_budget * doo_obj.eval_cost_single_point_normalised([1.0])
  opt_tree = OptTree(doo_obj, nu_max, rho_max, total_budget, K, C_init, tol, Randomize)
  results, index = opt_tree.run_PDOO(POO_mult)
  max_pt = doo_obj.get_unnormalised_coords(results[index][2])
  # IF not return history
  if not return_history:
    return results[index][0], max_pt, None
  history = Namespace()
  max_iter = int(total_budget)
  query_pts, query_vals = opt_tree.get_queried_pts()
  max_val = max(query_vals)
  history.query_step_idxs = [i for i in range(max_iter)]
  history.query_send_times = list(range(0, max_iter))
  history.query_receive_times = list(range(1, max_iter+1))
  history.query_points = [doo_obj.get_unnormalised_coords(x) for x in query_pts]
  history.query_vals = query_vals
  history.query_true_vals = query_vals
  history.curr_opt_vals = []
  history.curr_opt_points = []
  curr_max = -np.inf
  for idx, qv in enumerate(history.query_vals):
    if qv >= curr_max:
      curr_max = qv
      curr_opt_point = history.query_points[idx]
    history.curr_opt_vals.append(curr_max)
    history.curr_opt_points.append(curr_opt_point)
  history.query_eval_times = [1 for _ in range(max_iter)]
  history.curr_true_opt_vals = history.curr_opt_vals
  return max_val, max_pt, history


def pdoo_maximise_from_args(func, bounds, max_capital, *args, **kwargs):
  """ PDOO Maximise from Arguments. """
  doo_obj = DOOFunction(func, bounds)
  return pdoo_wrap(doo_obj, max_capital, *args, **kwargs)




"""
  A harness to load options.
  -- kandasamy@cs.cmu.edu
"""

import argparse
from copy import deepcopy

class LoadFromFile(argparse.Action):
  """ Argparse Action class to read from file. """
  def __call__(self, parser, namespace, values, option_string=None):
    with values as _file:
      options = []
      for line in _file.read().splitlines():
        if len(line) <= 2 or line[0:2] != '--':
          continue
        args_ = line.split()
        if len(args_) <= 1:
          continue
        options.extend(args_[0:2])
      parser.parse_args(options, namespace)


def get_option_specs(name, required=False, default=None, help_str='', **kwargs):
  """ A wrapper function to get a specification as a dictionary. """
  if isinstance(default, int):
    ret = {'name':name, 'required':required, 'default':default, 'help':help_str,
           'type':int}
  elif isinstance(default, float):
    ret = {'name':name, 'required':required, 'default':default, 'help':help_str,
           'type':float}
  else:
    ret = {'name':name, 'required':required, 'default':default, 'help':help_str}
  for key, value in list(kwargs.items()):
    ret[key] = value
  return ret


def _print_options(ondp, desc, reporter):
  """ Prints the options out. """
  if reporter is None:
    return
  title_str = 'Hyper-parameters for %s '%(desc)
  title_str = title_str + '-'*(80 - len(title_str))
  reporter.writeln(title_str)
  for key, value in sorted(ondp.items()):
    is_changed_str = '*' if value[0] != value[1] else ' '
    reporter.writeln('  %s %s %s'%(key.ljust(30), is_changed_str, str(value[1])))


def load_options(list_of_options, descr='Algorithm', reporter=None, cmd_line=False):
  """ Given a list of options, this reads them from the command line and returns
      a namespace with the values.
  """
  parser = argparse.ArgumentParser(description=descr)
  opt_names_default_parsed = {}
  for elem in list_of_options:
    opt_dict = deepcopy(elem)
    opt_name = opt_dict.pop('name')
    opt_names_default_parsed[opt_name] = [opt_dict['default'], None]
    if not opt_name.startswith('--'):
      opt_name = '--' + opt_name
    if opt_name == '--options':
      opt_dict['type'] = open
      opt_dict['action'] = LoadFromFile
    parser.add_argument(opt_name, **opt_dict)
  if cmd_line:
    args, _ = parser.parse_known_args()
  else:
    args = parser.parse_args(args=[])
  for key in opt_names_default_parsed:
    opt_names_default_parsed[key][1] = getattr(args, key)
  _print_options(opt_names_default_parsed, descr, reporter)
  return args




"""
  Unit tests for nn_examples.py
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=no-member
# pylint: disable=no-name-in-module
# pylint: disable=invalid-name

import os
import unittest
# Local imports
from . import nn_examples
from .nn_visualise import visualise_list_of_nns
from ..utils.base_test_class import BaseTestClass, execute_tests

# Test cases for nn_examples.py ----------------------------------------------------------
@unittest.skip
class NNExamplesTestCase(BaseTestClass):
  """ Unit test for some neural network examples. We are just testing for generation. """

  def __init__(self, *args, **kwargs):
    """ Constructor. """
    super(NNExamplesTestCase, self).__init__(*args, **kwargs)
    self.save_dir = '../scratch/unittest_examples'

  def test_vgg(self):
    """ Unit test for the VGG_net."""
    self.report('Testing the VGG net. ')
    nn_examples.get_vgg_net(2)

  def test_blocked_cnn(self):
    """ Unit test for a blocked CNN. """
    self.report('Testing a blocked CNN.')
    nn_examples.get_blocked_cnn(4, 4, 1)

  def test_generate_many_nns(self):
    """ Testing generation of many neural networks. """
    self.report('Testing generation of many NNs.')
    num_nns = 40
    cnns = nn_examples.generate_many_neural_networks('cnn', num_nns)
    visualise_list_of_nns(cnns, os.path.join(self.save_dir, 'cnn'))
    reg_mlps = nn_examples.generate_many_neural_networks('mlp-reg', num_nns)
    visualise_list_of_nns(reg_mlps, os.path.join(self.save_dir, 'reg_mlps'))
    class_mlps = nn_examples.generate_many_neural_networks('mlp-class', num_nns)
    visualise_list_of_nns(class_mlps, os.path.join(self.save_dir, 'class_mlps'))


if __name__ == '__main__':
  execute_tests()




"""
  A simple test for the fortran direct library.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name
# pylint: disable=unused-argument
# pylint: disable=relative-import

from __future__ import print_function
import numpy as np
try:
  import direct
  RUN_TESTS = True
except ImportError:
  RUN_TESTS = False


def main():
  """ Main function. """
  dim = 8
  obj = lambda x: (np.dot(x-0.1, x), 0)
  lower_bounds = [-1] * dim
  upper_bounds = [1] * dim
  dim = len(lower_bounds)
  eps = 1e-5
  max_func_evals = 5000
  max_iterations = max_func_evals
  algmethod = 0
  _log_file = 'log_dir_file'
  _kky_results_file = 'results_direct_file'
  fglobal = -1e100
  fglper = 0.01
  volper = -1.0
  sigmaper = -1.0

  def _objective_wrap(x, iidata, ddata, cdata, n, iisize, idsize, icsize):
    """ A wrapper to comply with the fortran requirements. """
    return obj(x)

  iidata = np.ones(0, dtype=np.int32)
  ddata = np.ones(0, dtype=np.float64)
  cdata = np.ones([0, 40], dtype=np.uint8)


  soln = direct.direct(_objective_wrap,
                       eps,
                       max_func_evals,
                       max_iterations,
                       np.array(lower_bounds, dtype=np.float64),
                       np.array(upper_bounds, dtype=np.float64),
                       algmethod,
                       _log_file,
                       _kky_results_file,
                       fglobal,
                       fglper,
                       volper,
                       sigmaper,
                       iidata,
                       ddata,
                       cdata
                      )
  print('Solution: %s'%(str(soln)))


if __name__ == '__main__':
  if RUN_TESTS:
    main()
  else:
    pass



"""
  Base Classes for probability distributions
  -- kvysyara@andrew.cmu.edu
"""

# pylint: disable=invalid-name

# Local imports
from ..sampling.slice import Slice
from ..sampling.metropolis import Metropolis, BinaryMetropolis
from ..sampling.nuts import NoUTurnSamplerDA as NUTS, LeapFrog


class Distribution(object):
  """ An abstract class which implements probability distributions """

  def __init__(self):
    """ Constructor. """
    self.domain = None
    self.dim = None
    super(Distribution, self).__init__()

  def draw_samples(self, method, *args):
    """ Returns random samples drawn from distribution. """
    raise NotImplementedError('Implement in a child class.')

  def draw_samples_random(self, method, *args):
    """ Returns random samples drawn from distribution. """
    raise NotImplementedError('Implement in a child class.')

  def get_mean(self):
    """ Returns the mean of the distribution. """
    raise NotImplementedError('Implement in a child class.')

  def get_variance(self):
    """ Returns the variance of the distribution. """
    raise NotImplementedError('Implement in a child class.')

  def get_domain(self):
    """ Returns domain """
    return self.domain

  def get_dim(self):
    """ Returns dimension """
    return self.dim


class Discrete(Distribution):
  """ Base class for Discrete distributions """

  def __init__(self):
    """ Constructor. """
    super(Discrete, self).__init__()

  def pmf(self, x):
    """ Returns pmf of distribution at x """
    raise NotImplementedError('Implement in a child class.')

  def draw_samples(self, method, size=None, init_sample=None, *args):
    """ Returns samples drawn from distribution. """
    if method == 'random':
      return self.draw_samples_random(size)
    elif method == 'metropolis':
      return self.draw_samples_metropolis(size, init_sample, *args)
    elif method == 'binarymetropolis':
      return self.draw_samples_binary_metropolis(size, init_sample, *args)

  def draw_samples_metropolis(self, size, init_sample, *args):
    """ Returns samples drawn from distribution using Metropolis sampler"""
    sampler = Metropolis(self, True, *args)
    return sampler.sample(init_sample, size)

  def draw_samples_binary_metropolis(self, size, init_sample, *args):
    """ Returns samples drawn from distribution using Binary Metropolis sampler"""
    sampler = BinaryMetropolis(self, True, *args)
    return sampler.sample(init_sample, size)


class Continuous(Distribution):
  """ Base class for Continuous distributions """

  def __init__(self):
    """ Constructor. """
    super(Continuous, self).__init__()

  def pdf(self, x):
    """ Returns pdf of distribution at x """
    raise NotImplementedError('Implement in a child class.')

  def draw_samples(self, method, size=None, init_sample=None, *args):
    """ Returns samples drawn from distribution. """
    if method == 'random':
      return self.draw_samples_random(size)
    elif method == 'nuts':
      return self.draw_samples_nuts(size, init_sample, *args)
    elif method == 'slice':
      return self.draw_samples_slice(size, init_sample, *args)
    elif method == 'metropolis':
      return self.draw_samples_metropolis(size, init_sample, *args)

  def draw_samples_slice(self, size, init_sample, *args):
    """ Returns samples drawn from distribution using Slice sampler"""
    sampler = Slice(self, *args)
    return sampler.sample(init_sample, size)

  def draw_samples_nuts(self, size, init_sample, *args):
    """ Returns samples drawn from distribution using NUTS sampler"""
    sampler = NUTS(self, LeapFrog)
    return sampler.sample(init_sample, size, *args)

  def draw_samples_metropolis(self, size, init_sample, *args):
    """ Returns samples drawn from distribution using Metropolis sampler"""
    sampler = Metropolis(self, *args)
    return sampler.sample(init_sample, size)




"""
  Implements the kernel, GP and fitter for multi-fidelity GPs.
  --kandasamy@cs.cmu.edu
"""
from __future__ import division

# pylint: disable=invalid-name
# pylint: disable=abstract-class-not-used

# Local imports
from . import kernel as gp_kernel
from .gp_core import GP, GPFitter
from ..utils.option_handler import load_options
from ..utils.reporters import get_reporter
from ..utils.ancillary_utils import get_list_of_floats_as_str


def get_ZX_from_ZZ_XX(ZZ, XX):
  """ Get a combined representation for the fidelity and domian data. """
  if hasattr(ZZ, '__iter__') and len(ZZ) == len(XX):
    return [(z, x) for (z, x) in zip(ZZ, XX)]
  else:
    return (ZZ, XX)


class MFGP(GP):
  """ A GP to be used in multi-fidelity settings. """

  def __init__(self, ZZ, XX, YY, mf_kernel,
               mean_func, noise_var, *args, **kwargs):
    """ Constructor. ZZ, XX, YY are the fidelity points, domain points and labels
        respectively.
        mf_kernel is either a combined kernel or a namespace with the following fields:
        {scale, fidel_kernel, domain_kernel}.
    """
    self.ZZ = list(ZZ)
    self.XX = list(XX)
    self.YY = list(YY)
    if not isinstance(mf_kernel, gp_kernel.Kernel):
      kernel = gp_kernel.CartesianProductKernel(mf_kernel.scale,
                 [mf_kernel.fidel_kernel, mf_kernel.domain_kernel])
      self.fidel_kernel = mf_kernel.fidel_kernel
      self.domain_kernel = mf_kernel.domain_kernel
    else:
      kernel = mf_kernel
    ZX = self.get_ZX_from_ZZ_XX(ZZ, XX) # The 'X' data
    # Call super constructor
    super(MFGP, self).__init__(ZX, YY, kernel, mean_func, noise_var, *args, **kwargs)

  @classmethod
  def get_ZX_from_ZZ_XX(cls, ZZ, XX):
    """ Get a combined representation for the fidelity and domian data.
        Can be overridden by a child class if there is a more efficient representation."""
    return get_ZX_from_ZZ_XX(ZZ, XX)

  def eval_at_fidel(self, ZZ_test, XX_test, *args, **kwargs):
    """ Evaluates the GP at [ZZ_test, XX_test]. Read eval in gp_core.GP for more details.
    """
    ZX_test = self.get_ZX_from_ZZ_XX(ZZ_test, XX_test)
    return self.eval(ZX_test, *args, **kwargs)

  def eval_at_fidel_with_hallucinated_observations(self, ZZ_test, XX_test,
                                                   ZZ_halluc, XX_halluc, *args, **kwargs):
    """ Evaluates with hallucinated observations. """
    ZX_test = self.get_ZX_from_ZZ_XX(ZZ_test, XX_test)
    ZX_halluc = self.get_ZX_from_ZZ_XX(ZZ_halluc, XX_halluc)
    return self.eval_with_hallucinated_observations(ZX_test, ZX_halluc, *args, **kwargs)

  def set_mf_data(self, ZZ, XX, YY, build_posterior=True):
    """ Sets the MF data for the GP. """
    self.ZZ = list(ZZ)
    self.XX = list(XX)
    self.YY = list(YY)
    ZX = self.get_ZX_from_ZZ_XX(ZZ, XX) # The 'X' data
    super(MFGP, self).set_data(ZX, YY, build_posterior)

  def add_mf_data_multiple(self, ZZ_new, XX_new, YY_new, *args, **kwargs):
    """ Adds new data to the multi-fidelity GP. """
    ZX_new = self.get_ZX_from_ZZ_XX(ZZ_new, XX_new)
    self.ZZ.extend(ZZ_new)
    self.XX.extend(XX_new)
    self.add_data_multiple(ZX_new, YY_new, *args, **kwargs)

  def add_mf_data_single(self, zz_new, xx_new, yy_new, *args, **kwargs):
    """ Adds a single new data to the multi-fidelity GP. """
    self.add_mf_data_multiple([zz_new], [xx_new], [yy_new], *args, **kwargs)

  def draw_mf_samples(self, num_samples, ZZ_test=None, XX_test=None, *args, **kwargs):
    """ Draws samples from a multi-fidelity GP. """
    ZX_test = None if ZZ_test is None else self.get_ZX_from_ZZ_XX(ZZ_test, XX_test)
    return self.draw_samples(num_samples, ZX_test, *args, **kwargs)

  def get_fidel_kernel(self):
    """ Return the fidel_space kernel. """
    return self.fidel_kernel

  def get_domain_kernel(self):
    """ Return the domain kernel. """
    return self.domain_kernel

  def _child_str(self):
    """ Returns a string representation of the MF-GP. """
    if hasattr(self, 'fidel_kernel') and hasattr(self, 'domain_kernel'):
      fidel_ke_str = self._get_kernel_str(self.fidel_kernel)
      domain_ke_str = self._get_kernel_str(self.domain_kernel)
      kernel_str = 'fid:: %s, dom:: %s'%(fidel_ke_str, domain_ke_str)
    else:
      kernel_str = str(self.kernel)
    ret = 'scale: %0.3f, %s'%(self.kernel.hyperparams['scale'], kernel_str)
    return ret

  @classmethod
  def _get_kernel_str(cls, kern):
    """ Gets a string format of the kernel depending on whether it is SE/Poly. """
    if isinstance(kern, gp_kernel.ExpDecayKernel):
      ret = 'expd: offs=%0.3f, pow=%s'%(kern.hyperparams['offset'],
        get_list_of_floats_as_str(kern.hyperparams['powers']))
    elif isinstance(kern, gp_kernel.SEKernel) or isinstance(kern, gp_kernel.MaternKernel):
      hp_name = 'dim_bandwidths'
      kern_name = 'se' if isinstance(kern, gp_kernel.SEKernel) else \
                  'matern(%0.1f)'%(kern.hyperparams['nu'])
      if kern.dim > 4:
        ret = '%0.4f(avg)'%(kern.hyperparams[hp_name].mean())
      else:
        ret = get_list_of_floats_as_str(kern.hyperparams[hp_name])
      ret = kern_name + ': ' + ret
    elif isinstance(kern, gp_kernel.PolyKernel):
      ret = 'poly: %s'%(get_list_of_floats_as_str(kern.hyperparams['dim_scalings']))
    else:    # Return an empty string.
      ret = str(kern)
    return ret


class MFGPFitter(GPFitter):
  """ A GP Fitter for Multi-fidelity GPs. This is mostly a wrapper for MFGPs that
      want to use GPFitter. All the heavy lifting is happening in GPFitter. """
  # pylint: disable=abstract-method

  def __init__(self, ZZ, XX, YY, options=None, reporter=None):
    """ Constructor. """
    reporter = get_reporter(reporter)
    if isinstance(options, list):
      options = load_options(options, 'MFGP', reporter=self.reporter)
    self.ZZ = ZZ
    self.XX = XX
    self.YY = YY
    self.num_tr_data = len(self.YY)
    ZX = get_ZX_from_ZZ_XX(ZZ, XX)
    super(MFGPFitter, self).__init__(ZX, YY, options, reporter)




"""
  Test cases for functions in general_utils.py
  -- kandasamy@cs.cmu.edu
"""
from __future__ import absolute_import
from __future__ import division

# pylint: disable=import-error
# pylint: disable=no-member
# pylint: disable=invalid-name
# pylint: disable=relative-import

import numpy as np
from . import general_utils
from .base_test_class import BaseTestClass, execute_tests


class GeneralUtilsTestCase(BaseTestClass):
  """Unit test class for general utilities. """

  def __init__(self, *args, **kwargs):
    super(GeneralUtilsTestCase, self).__init__(*args, **kwargs)

  def setUp(self):
    """ Sets up attributes. """
    # For dist squared
    self.X1 = np.array([[1, 2, 3], [1, 2, 4], [2, 3, 4.5]])
    self.X2 = np.array([[1, 2, 4], [1, 2, 5], [2, 3, 5]])
    self.true_dist_sq = np.array([[1, 4, 6], [0, 1, 3], [2.25, 2.25, 0.25]])

  def test_dist_squared(self):
    """ Tests the squared distance function. """
    self.report('dist_squared')
    comp_dist_sq = general_utils.dist_squared(self.X1, self.X2)
    assert (self.true_dist_sq == comp_dist_sq).all()

  def test_mapping_to_cube_and_bound(self):
    """ Test map_to_cube and map_to_bounds. """
    self.report('map_to_cube and map_to_bounds')
    bounds = np.array([[1, 3], [2, 4], [5, 6]])
    x = np.array([1.7, 3.1, 5.5])
    X = np.array([[1.7, 3.1, 5.5], [2.1, 2.9, 5.0]])
    y = np.array([0.35, 0.55, 0.5])
    Y = np.array([[0.35, 0.55, 0.5], [0.55, 0.45, 0]])
    # Map to cube
    y_ = general_utils.map_to_cube(x, bounds)
    Y_ = general_utils.map_to_cube(X, bounds)
    # Map to Bounds
    x_ = general_utils.map_to_bounds(y, bounds)
    X_ = general_utils.map_to_bounds(Y, bounds)
    # Check if correct.
    assert np.linalg.norm(y - y_) < 1e-5
    assert np.linalg.norm(Y - Y_) < 1e-5
    assert np.linalg.norm(x - x_) < 1e-5
    assert np.linalg.norm(X - X_) < 1e-5

  def test_compute_average_sq_prediction_error(self):
    """ Tests compute_average_sq_prediction_error. """
    self.report('compute_average_sq_prediction_error')
    Y1 = [0, 1, 2]
    Y2 = [2, 0, 1]
    res = general_utils.compute_average_sq_prediction_error(Y1, Y2)
    assert np.abs(res - 2.0) < 1e-5

  def test_stable_cholesky(self):
    """ Tests for stable cholesky. """
    self.report('stable_cholesky')
    M = np.random.normal(size=(5, 5))
    M = M.dot(M.T)
    L = general_utils.stable_cholesky(M)
    assert np.linalg.norm(L.dot(L.T) - M) < 1e-5

  def test_draw_gaussian_samples(self):
    """ Tests for draw gaussian samples. """
    self.report('draw_gaussian_samples. Probabilistic test, could fail at times')
    num_samples = 10000
    num_pts = 3
    mu = list(range(num_pts))
    K = np.random.normal(size=(num_pts, num_pts))
    K = K.dot(K.T)
    samples = general_utils.draw_gaussian_samples(num_samples, mu, K)
    sample_mean = samples.mean(axis=0)
    sample_centralised = samples - sample_mean
    sample_covar = sample_centralised.T.dot(sample_centralised) / num_samples
    mean_tol = 4 * np.linalg.norm(mu) / np.sqrt(num_samples)
    covar_tol = 4 * np.linalg.norm(K) / np.sqrt(num_samples)
    mean_err = np.linalg.norm(mu - sample_mean)
    covar_err = np.linalg.norm(K - sample_covar)
    self.report('Mean error (tol): ' + str(mean_err) + ' (' + str(mean_tol) + ')',
                'test_result')
    self.report('Cov error (tol): ' + str(covar_err) + ' (' + str(covar_tol) + ')',
                'test_result')
    assert mean_err < mean_tol
    assert covar_err < covar_tol

  def test_ordering_and_reordering(self):
    """ Tests for ordering and reordering of lists. """
    self.report('Testing ordering and reordering of lists.')
    # Test 1
    orig_list = ['k', 'i', 'r', 't', 'h']
    ordering = [3, 1, 0, 4, 2]
    reordered_list = general_utils.reorder_list(orig_list, ordering)
    rereordered_list = general_utils.get_original_order_from_reordered_list(
                         reordered_list, ordering)
    assert reordered_list == ['t', 'i', 'k', 'h', 'r']
    assert rereordered_list == orig_list
    # Test 2
    N = 20
    orig_list = list(np.random.random(N,))
    ordering = np.random.permutation(N)
    reordered_list = general_utils.reorder_list(orig_list, ordering)
    rereordered_list = general_utils.get_original_order_from_reordered_list(
                         reordered_list, ordering)
    assert rereordered_list == orig_list

  def test_list_flattening(self):
    """ Tests list flattening. """
    self.report('Testing list flattening.')
    # Test 1
    test_lists = [[4, 5, 1],
                  [4, 5, 6, ['k', None], 5]]
    test_results = [[4, 5, 1],
                    [4, 5, 6, 'k', None, 5]]
    for tl, tr in zip(test_lists, test_results):
      result = general_utils.flatten_list_of_objects_and_lists(tl)
      assert tr == result
    # Test 2
    test_lists.extend([['q', 'we', [7, ['3', str, int], 9], 'k']])
    test_results.extend([['q', 'we', 7, '3', str, int, 9, 'k']])
    for tl, tr in zip(test_lists, test_results):
      result = general_utils.flatten_nested_lists(tl)
      assert tr == result

  def test_pairwise_hamming_kernel(self):
    """ Tests pairwise hamming distances. """
    self.report('Testing hamming distance.')
    # Test 1
    test_data = [# data 1
                ([[4, 5], ['kky', 5], ['s', None]],
                 [[2, 5], ['s', None], ['kky', None], ['s', 1]],
                 [0.4, 0.6],
                 [[1.0, 0.6, 0], [0.6, 1.0, 0], [0, 0, 1.0]],
                 [[1, 0, 0, 0], [0, 1, 0.6, 0.4], [0, 0.6, 1, 0], [0, 0.4, 0, 1]],
                 [[0.6, 0, 0, 0], [0.6, 0, 0.4, 0], [0, 1, 0.6, 0.4]]),
                 # data 2
                ([[4, 5, 6], ['kky', 5, 'cat']],
                 [[2, 5, 6], ['s', None, 'cat'], ['kky', None, 6]],
                 None,
                 [[1.0, 1/3.0], [1/3.0, 1]],
                 [[1.0, 0, 1/3.0], [0, 1, 1/3.0], [1/3.0, 1/3.0, 1]],
                 [[2/3.0, 0, 1/3.0], [1/3.0, 1/3.0, 1/3.0]]),
                ]
    for td in test_data:
      res_00 = np.array(td[-3])
      res_11 = np.array(td[-2])
      res_01 = np.array(td[-1])
      res_10 = res_01.T
      kern_00 = general_utils.pairwise_hamming_kernel(td[0], td[0], td[2])
      kern_11 = general_utils.pairwise_hamming_kernel(td[1], td[1], td[2])
      kern_01 = general_utils.pairwise_hamming_kernel(td[0], td[1], td[2])
      kern_10 = general_utils.pairwise_hamming_kernel(td[1], td[0], td[2])
#       print res_00
#       print kern_00
#       print res_11
#       print kern_11
#       print res_01
#       print kern_01
#       print res_10
#       print kern_10
      assert np.linalg.norm(res_00 - kern_00) < 1e-5
      assert np.linalg.norm(res_11 - kern_11) < 1e-5
      assert np.linalg.norm(res_01 - kern_01) < 1e-5
      assert np.linalg.norm(res_10 - kern_10) < 1e-5


if __name__ == '__main__':
  execute_tests()




"""
  Harness for calling experiments including Multi-fidelity.
  -- kandasamy@cs.cmu.edu
"""
# Convention: fidel/fidel_space always comes before point/domain --KK

# pylint: disable=abstract-class-little-used
# pylint: disable=abstract-method
# pylint: disable=invalid-name
# pylint: disable=too-many-arguments

from __future__ import print_function

from argparse import Namespace
import numpy as np
import pickle
# Local imports
from .cp_domain_utils import load_config_file, sample_from_cp_domain, \
                                get_processed_func_from_raw_func_via_config, \
                                get_processed_func_from_raw_func_for_cp_domain
from .domains import EuclideanDomain
from .exd_utils import EVAL_ERROR_CODE
from ..utils.general_utils import map_to_cube, map_to_bounds

_FIDEL_TOL = 1e-2


class CalledMultiFidelOnSingleFidelCaller(Exception):
  """ An exception to handle calling multi-fidelity experiments on single-fidelity
      callers.
  """
  def __init__(self, exp_caller):
    """ Constructor. """
    err_msg = ('ExperimentCaller %s is not a multi-fidelity caller. ' + \
               'Please use eval_single or eval_multiple.')%(str(exp_caller))
    super(CalledMultiFidelOnSingleFidelCaller, self).__init__(err_msg)


# Base class for Experiments ============================================================
class ExperimentCaller(object):
  """ Function Caller class. """

  # Methods needed for construction and set up ------------------------------------------
  def __init__(self, experiment, domain, descr='',
               noise_type='no_noise', noise_scale=None,
               fidel_space=None, fidel_cost_func=None, fidel_to_opt=None):
    """ Constructor. """
    self.experiment = experiment
    self.domain = domain
    self.descr = descr
    self._set_up_noise(noise_type, noise_scale)
    self._mf_set_up(fidel_space, fidel_cost_func, fidel_to_opt)

  def _set_up_noise(self, noise_type, noise_scale):
    """ Sets up noise. """
    self.noise_type = noise_type
    if noise_type == 'no_noise':
      self._is_noisy = False
      self.noise_scale = None
      self.noise_adder_single = None
      self.noise_adder_multiple = None
    else:
      self._is_noisy = True
      self.noise_scale = noise_scale
      self.noise_type = noise_type
      self._set_up_noisy_evals()

  def _set_up_noisy_evals(self):
    """ Set up noisy evaluations. """
    raise NotImplementedError('Implement this in a noisy evaluator.')

  def _mf_set_up(self, fidel_space, fidel_cost_func, fidel_to_opt):
    """ Sets up multi-fidelity. """
    if any([elem is None for elem in [fidel_space, fidel_cost_func, fidel_to_opt]]):
      # If any of these are None, assert that all of them are. Otherwise through an
      # error.
      try:
        assert fidel_space is None
        assert fidel_cost_func is None
        assert fidel_to_opt is None
      except AssertionError:
        raise ValueError('Either all fidel_space, fidel_cost_func, and fidel_to_opt' +
                         'should be None or should be not-None')
      self._is_mf = False
    else:
      self.fidel_space = fidel_space
      self.fidel_cost_func = fidel_cost_func
      self.fidel_to_opt = fidel_to_opt
      self._is_mf = True

  def is_noisy(self):
    """ Returns true if noisy. """
    return self._is_noisy

  def is_mf(self):
    """ Returns True if this is multi-fidelity. """
    return self._is_mf

  def is_fidel_to_opt(self, fidel):
    """ Returns True if fidel is the fidel_to_opt. Naively tests for equality, but can
        be overridden by a child class. """
    if not self.is_mf():
      raise CalledMultiFidelOnSingleFidelCaller()
    return self.fidel_space.members_are_equal(fidel, self.fidel_to_opt)

  # Evaluation --------------------------------------------------------------------------
  def get_noisy_value(self, true_val):
    """ Gets the noisy evaluation. """
    raise NotImplementedError('Implement in a child class.')

  @classmethod
  def _pickle_dump_to_result_file(cls, qinfo, file_handle):
    """ Writes to the pickle file. """
    pickle.dump(qinfo, file_handle)

  def _eval_single_common_wrap_up(self, true_val, qinfo, noisy, caller_eval_cost):
    """ Wraps up the evaluation by adding noise and adding info to qinfo.
        Common to both single and mutli-fidelity eval experiments.
    """
    if true_val == EVAL_ERROR_CODE:
      val = EVAL_ERROR_CODE
    elif noisy and self.is_noisy():
      val = self.get_noisy_value(true_val)
    else:
      val = true_val
    # put everything into qinfo
    qinfo = Namespace() if qinfo is None else qinfo
    qinfo.true_val = true_val
    qinfo.val = val
    qinfo.caller_eval_cost = caller_eval_cost
    # Writing to a file
    if hasattr(qinfo, 'result_file') and qinfo.result_file is not None:
      if isinstance(qinfo.result_file, str):
        file_handle = open(qinfo.result_file, 'wb')
        self._pickle_dump_to_result_file(qinfo, file_handle)
        file_handle.close()
      elif isinstance(qinfo.result_file, file):
        self._pickle_dump_to_result_file(qinfo, qinfo.result_file)
      else:
        raise ValueError('qinfo.result_file should a string or file object.')
    return val, qinfo

  def _get_true_val_from_experiment_at_point(self, point):
    """ Returns the true value from the experiment. Can be overridden by child class if
        experiment is represented differently. """
    assert self.domain.is_a_member(point)
    return self.experiment(point)

  def _get_true_val_from_experiment_at_fidel_point(self, fidel, point):
    """ Returns the true value from the experiment. Can be overridden by child class if
        experiment is represented differently. """
    assert self.fidel_space.is_a_member(fidel)
    assert self.domain.is_a_member(point)
    return self.experiment(fidel, point)

  # Single fidelity evaluations --------------------------------------------
  def eval_single(self, point, qinfo=None, noisy=True):
    """ Evaluates experiment at a single point point. If the experiment_caller is noisy by
        default, we can obtain a noiseless evaluation by setting noisy to be False.
    """
    if self.is_mf(): # if multi-fidelity call the experiment at fidel_to_opt
      return self.eval_at_fidel_single(self.fidel_to_opt, point, qinfo, noisy)
    else:
      if qinfo is None:
        qinfo = Namespace()
      true_val = self._get_true_val_from_experiment_at_point(point)
      qinfo.point = point
      val, qinfo = self._eval_single_common_wrap_up(true_val, qinfo, noisy, None)
      return val, qinfo

  def eval_multiple(self, points, qinfos=None, noisy=True):
    """ Evaluates multiple points. If the experiment_caller is noisy by
        default, we can obtain a noiseless evaluation by setting noisy to be False.
    """
    qinfos = [None] * len(points) if qinfos is None else qinfos
    ret_vals = []
    ret_qinfos = []
    for i in range(len(points)):
      val, qinfo = self.eval_single(points[i], qinfos[i], noisy)
      ret_vals.append(val)
      ret_qinfos.append(qinfo)
    return ret_vals, ret_qinfos

  # Multi-fidelity evaluations -------------------------------------------------
  def eval_at_fidel_single(self, fidel, point, qinfo=None, noisy=True):
    """ Evaluates experiment at a single (fidel, point). If the experiment_caller is
        noisy by default, we can obtain a noiseless evaluation by setting noisy to be
        False.
    """
    if not self.is_mf():
      raise CalledMultiFidelOnSingleFidelCaller(self)
    if qinfo is None:
      qinfo = Namespace()
    true_val = self._get_true_val_from_experiment_at_fidel_point(fidel, point)
    cost_at_fidel = self.fidel_cost_func(fidel)
    qinfo.fidel = fidel
    qinfo.point = point
    qinfo.cost_at_fidel = cost_at_fidel
    val, qinfo = self._eval_single_common_wrap_up(true_val, qinfo, noisy,
                                                  cost_at_fidel)
    return val, qinfo

  def eval_at_fidel_multiple(self, fidels, points, qinfos=None, noisy=True):
    """ Evaluates experiment at a multiple (fidel, point) pairs.
        If the experiment_caller is noisy by
        default, we can obtain a noiseless evaluation by setting noisy to be False.
    """
    qinfos = [None] * len(points) if qinfos is None else qinfos
    ret_vals = []
    ret_qinfos = []
    for i in range(len(points)):
      val, qinfo = self.eval_at_fidel_single(fidels[i], points[i], qinfos[i], noisy)
      ret_vals.append(val)
      ret_qinfos.append(qinfo)
    return ret_vals, ret_qinfos

  # Eval from qinfo
  def eval_from_qinfo(self, qinfo, *args, **kwargs):
    """ Evaluates from a qinfo object. Returns the qinfo. """
    if not hasattr(qinfo, 'fidel'):
      _, qinfo = self.eval_single(qinfo.point, qinfo, *args, **kwargs)
      return qinfo
    else:
      _, qinfo = self.eval_at_fidel_single(qinfo.fidel, qinfo.point, qinfo,
                                            *args, **kwargs)
      return qinfo

  # Cost -----------------------------------------------------------------------------
  def _get_true_cost_from_fidel_cost_func_at_fidel(self, fidel):
    """ Returns the true value from the experiment. Can be overridden by child class if
        fidel_cost_func is represented differently. """
    return float(self.fidel_cost_func(fidel))

  def cost_single(self, fidel):
    """ Returns the cost at a single fidelity. """
    return self._get_true_cost_from_fidel_cost_func_at_fidel(fidel)

  def cost_multiple(self, fidels):
    """ Returns the cost at multiple fidelities. """
    return [self._get_true_cost_from_fidel_cost_func_at_fidel(fidel) for fidel in fidels]

  def cost_ratio_single(self, fidel_numerator, fidel_denominator=None):
    """ Returns the cost ratio. If fidel_denominator is None, we set it to be
        fidel_to_opt.
    """
    if fidel_denominator is None:
      fidel_denominator = self.fidel_to_opt
    return self.cost_single(fidel_numerator) / self.cost_single(fidel_denominator)

  def cost_ratio_multiple(self, fidels_numerator, fidel_denominator=None):
    """ Returns the cost ratio for multiple fidels. If fidel_denominator is None,
        we set it to be fidel_to_opt.
    """
    if fidel_denominator is None:
      fidel_denominator = self.fidel_to_opt
    numerator_costs = self.cost_multiple(fidels_numerator)
    denom_cost = self.cost_single(fidel_denominator)
    ret = [x/denom_cost for x in numerator_costs]
    return ret

  # Other methods ---------------------------------------------------------------------
  def get_candidate_fidels(self, domain_point, filter_by_cost=True, *args, **kwargs):
    """ Returns candidate fidelities at domain_point.
        If filter_by_cost is true returns only those for which cost is larger than
        fidel_to_opt.
    """
    if not self.is_mf():
      raise CalledMultiFidelOnSingleFidelCaller(self)
    candidate_fidels = self._child_get_candidate_fidels(domain_point, filter_by_cost,
                                                        *args, **kwargs)
    candidate_fidels.append(self.fidel_to_opt)
    return candidate_fidels

  def _child_get_candidate_fidels(self, domain_point, filter_by_cost=True,
                                  *args, **kwargs):
    """ Returns candidate fidelities at domain_point for the child class.
        If filter_by_cost is true returns only those for which cost is larger than
        fidel_to_opt.
    """
    raise NotImplementedError('Implement in a child class.')

  def get_candidate_fidels_and_cost_ratios(self, domain_point, filter_by_cost=True,
                                           *args, **kwargs):
    """ Returns candidate fidelities and the cost ratios.
        If filter_by_cost is true returns only those for which cost is larger than
        fidel_to_opt.
    """
    candidates = self.get_candidate_fidels(domain_point, filter_by_cost=False,
                                           add_fidel_to_opt=False, *args, **kwargs)
    fidel_cost_ratios = self.cost_ratio_multiple(candidates)
    if filter_by_cost:
      filtered_idxs = np.where(np.array(fidel_cost_ratios) < 1.0)[0]
      candidates = [candidates[idx] for idx in filtered_idxs]
      fidel_cost_ratios = [fidel_cost_ratios[idx] for idx in filtered_idxs]
      # But re-add fidel_to_opt.
      candidates.append(self.fidel_to_opt)
      fidel_cost_ratios.append(1.0)
    return candidates, fidel_cost_ratios

  def get_information_gap(self, fidels):
    """ Returns the information gap w.r.t fidel_to_opt. """
    raise NotImplementedError('Implement in a child class.')


# A class for evaluation multiple functions ==========================================
# That is, the output of each function is a real number.
class MultiFunctionCaller(ExperimentCaller):
  """ An Experiment Caller specifically for evaluating a collection of functions. """

  def __init__(self, funcs, domain, descr='',
               argmax=None, maxval=None, argmin=None, minval=None,
               noise_type='no_noise', noise_scale=None,
               fidel_space=None, fidel_cost_func=None, fidel_to_opt=None):
    """ Constructor. """
    self.funcs = funcs
    self.argmax = argmax
    self.maxval = maxval
    self.argmin = argmin
    self.minval = minval
    experiment_caller = self._get_experiment_caller()
    super(MultiFunctionCaller, self).__init__(experiment_caller, domain, descr,
      noise_type=noise_type, noise_scale=noise_scale,
      fidel_space=fidel_space, fidel_cost_func=fidel_cost_func, fidel_to_opt=fidel_to_opt)

  def _get_experiment_caller(self):
    """ Returns the experiment caller. If funcs is a list, then the function caller
        returns a list of element for each function in funcs.
        Otherwise, it just returns self.funcs which is assumed to be callable.
    """
    if isinstance(self.funcs, list):
      self._has_many_functions = True
      self.num_funcs = len(self.funcs)
      return lambda *_args, **_kwargs: [float(f(*_args, **_kwargs)) for f in self.funcs]
    elif isinstance(self.funcs, tuple) and len(self.funcs) == 2:
      self._has_many_functions = True
      self.num_funcs = self.funcs[1]
#       return lambda *_args, **_kwargs: self.funcs[0](*_args, **_kwargs)
      return self.funcs[0]
    else:
      self._has_many_functions = False
      return lambda *_args, **_kwargs: float(self.funcs(*_args, **_kwargs))

  def _set_up_noisy_evals(self):
    """ Sets up noisy evaluations. """
    if self.noise_type == 'gauss':
      self.noise_adder_single = lambda: self.noise_scale * np.random.normal()
      self.noise_adder_multiple = \
        lambda num_samples: self.noise_scale * np.random.normal((num_samples,))
    elif self.noise_type == 'uniform':
      self.noise_adder_single = lambda: self.noise_scale * (np.random.random() - 0.5)
      self.noise_adder_multiple = \
        lambda num_samples: self.noise_scale * (np.random.normal((num_samples,)) - 0.5)
    else:
      raise NotImplementedError(('Not implemented %s noise yet')%(self.noise_type))

  def get_noisy_value(self, true_val):
    """ Gets the noisy evaluation. """
    if self._has_many_functions:
      return [tv + self.noise_adder_single() for tv in true_val]
    else:
      return true_val + self.noise_adder_single()


# A function caller where the domain is Euclidean ======================================
class EuclideanMultiFunctionCaller(MultiFunctionCaller):
  """ An experiment caller on Euclidean spaces. """

  def __init__(self, funcs, raw_domain, descr='', vectorised=False,
               to_normalise_domain=True,
               raw_argmax=None, maxval=None, raw_argmin=None, minval=None,
               noise_type='no_noise', noise_scale=None,
               raw_fidel_space=None, fidel_cost_func=None, raw_fidel_to_opt=None):
    """ Constructor. """
    # pylint: disable=too-many-arguments
    # Prelims
    if hasattr(raw_domain, '__iter__'):
      raw_domain = EuclideanDomain(raw_domain)
    if hasattr(raw_fidel_space, '__iter__'):
      raw_fidel_space = EuclideanDomain(raw_fidel_space)
    self.vectorised = vectorised
    self.domain_is_normalised = to_normalise_domain
    # Set domain and and argmax/argmin
    self.raw_domain = raw_domain
    self.raw_argmax = raw_argmax
    argmax = None if raw_argmax is None else self.get_normalised_domain_coords(raw_argmax)
    self.raw_argmin = raw_argmin
    argmin = None if raw_argmin is None else self.get_normalised_domain_coords(raw_argmin)
    domain = EuclideanDomain([[0, 1]] * raw_domain.dim) if to_normalise_domain \
             else raw_domain
    # Set fidel_space
    if raw_fidel_space is not None:
      self.raw_fidel_space = raw_fidel_space
      self.raw_fidel_to_opt = raw_fidel_to_opt
      fidel_space = EuclideanDomain([[0, 1]] * raw_fidel_space.dim) \
                    if to_normalise_domain else raw_fidel_space
      fidel_to_opt = self.get_normalised_fidel_coords(raw_fidel_to_opt)
      self.fidel_space_diam = np.linalg.norm(
        fidel_space.bounds[:, 1] - fidel_space.bounds[:, 0])
    else:
      fidel_space = None
      fidel_to_opt = None
    # Now call the super constructor
    super(EuclideanMultiFunctionCaller, self).__init__(funcs=funcs, domain=domain,
                 descr=descr, argmax=argmax, maxval=maxval, argmin=argmin, minval=minval,
                 noise_type=noise_type, noise_scale=noise_scale,
                 fidel_space=fidel_space, fidel_cost_func=fidel_cost_func,
                 fidel_to_opt=fidel_to_opt)

  def is_fidel_to_opt(self, fidel):
    """ Returns True if fidel is the fidel_to_opt. """
    return np.linalg.norm(fidel - self.fidel_to_opt) < _FIDEL_TOL * self.fidel_space_diam

  # Methods required for normalising coordinates -----------------------------------------
  def get_normalised_fidel_coords(self, Z):
    """ Maps points in the original fidelity space to the cube. """
    if self.domain_is_normalised:
      return map_to_cube(Z, self.raw_fidel_space.bounds)
    else:
      return Z

  def get_normalised_domain_coords(self, X):
    """ Maps points in the original domain to the cube. """
    if self.domain_is_normalised:
      return map_to_cube(X, self.raw_domain.bounds)
    else:
      return X

  def get_normalised_fidel_domain_coords(self, Z, X):
    """ Maps points in the original space to the cube. """
    ret_Z = None if Z is None else self.get_normalised_fidel_coords(Z)
    ret_X = None if X is None else self.get_normalised_domain_coords(X)
    return ret_Z, ret_X

  def get_raw_fidel_coords(self, Z):
    """ Maps points from the fidelity space cube to the original space. """
    if self.domain_is_normalised:
      return map_to_bounds(Z, self.raw_fidel_space.bounds)
    else:
      return Z

  def get_raw_domain_coords(self, X):
    """ Maps points from the domain cube to the original space. """
    if self.domain_is_normalised:
      return map_to_bounds(X, self.raw_domain.bounds)
    else:
      return X

  def get_raw_fidel_domain_coords(self, Z, X):
    """ Maps points from the cube to the original spaces. """
    ret_Z = None if Z is None else self.get_raw_fidel_coords(Z)
    ret_X = None if X is None else self.get_raw_domain_coords(X)
    return ret_Z, ret_X

  # Override _get_true_val_from_experiment_at_point and
  # _get_true_val_from_experiment_at_fidel_point
  # so as to account for normalisation of the domain and/or fidel_space
  def _get_true_val_from_experiment_at_point(self, point):
    """ Evaluates experiment by first unnormalising point. """
    raw_dom_coords = self.get_raw_domain_coords(point)
    assert self.raw_domain.is_a_member(raw_dom_coords)
    if self.vectorised:
      raw_dom_coords = raw_dom_coords.reshape((-1, 1))
    return self.experiment(raw_dom_coords)

  def _get_true_val_from_experiment_at_fidel_point(self, fidel, point):
    """ Evaluates experiment by first unnormalising point. """
    raw_fidel_coords = self.get_raw_fidel_coords(fidel)
    assert self.raw_fidel_space.is_a_member(raw_fidel_coords)
    raw_dom_coords = self.get_raw_domain_coords(point)
    assert self.raw_domain.is_a_member(raw_dom_coords)
    if self.vectorised:
      raw_dom_coords = raw_dom_coords.reshape((-1, 1))
      raw_fidel_coords = raw_fidel_coords.reshape((-1, 1))
    return self.experiment(self.get_raw_fidel_coords(fidel),
                           self.get_raw_domain_coords(point))

  def _get_true_cost_from_fidel_cost_func_at_fidel(self, fidel):
    """ Evaluates fidel_cost_func by unnormalising fidel. """
    raw_fidel_coords = self.get_raw_fidel_coords(fidel)
    assert self.raw_fidel_space.is_a_member(raw_fidel_coords)
    if self.vectorised:
      raw_fidel_coords = raw_fidel_coords.reshape((-1, 1))
    return float(self.fidel_cost_func(raw_fidel_coords))

  def _child_get_candidate_fidels(self, domain_point, filter_by_cost=True,
                                  *args, **kwargs):
    """ Returns candidate fidelities at domain_point.
        If filter_by_cost is true returns only those for which cost is larger than
        fidel_to_opt.
    """
    if self.fidel_space.dim == 1:
      norm_candidates = np.linspace(0, 1, 100).reshape((-1, 1))
    elif self.fidel_space.dim == 2:
      num_per_dim = 25
      norm_candidates = (np.indices((num_per_dim, num_per_dim)).reshape(2, -1).T + 0.5) \
                        / float(num_per_dim)
    elif self.fidel_space.dim == 3:
      num_per_dim = 10
      cand_1 = (np.indices((num_per_dim, num_per_dim, num_per_dim)).reshape(3, -1).T
                + 0.5) / float(num_per_dim)
      cand_2 = np.random.random((1000, self.fidel_space.dim))
      norm_candidates = np.vstack((cand_1, cand_2))
    else:
      norm_candidates = np.random.random((4000, self.fidel_space.dim))
    # Now unnormalise if necessary
    if self.domain_is_normalised:
      candidates = norm_candidates
    else:
      candidates = map_to_bounds(candidates, self.fidel_space.bounds)
    if filter_by_cost:
      fidel_costs = self.cost_multiple(candidates)
      filtered_idxs = np.where(np.array(fidel_costs) <
                               self.cost_single(self.fidel_to_opt))[0]
      candidates = candidates[filtered_idxs, :]
    # Finally, always add the highest fidelity
    candidates = list(candidates)
    return candidates

  def get_information_gap(self, fidels):
    """ Returns distances to fidel_to_opt. """
    if not self.is_mf():
      raise CalledMultiFidelOnSingleFidelCaller(self)
    return [np.linalg.norm(fidel - self.fidel_to_opt)/self.fidel_space_diam \
            for fidel in fidels]


class CPMultiFunctionCaller(MultiFunctionCaller):
  """ Cartesian product multi-function caller. """

  def __init__(self, funcs, domain, descr='', raw_funcs=None, domain_orderings=None,
               argmax=None, maxval=None, argmin=None, minval=None,
               noise_type='no_noise', noise_scale=None,
               fidel_space=None, fidel_cost_func=None, fidel_to_opt=None,
               fidel_space_orderings=None):
    """ Constructor. """
    self.raw_funcs = raw_funcs
    self.domain_orderings = domain_orderings
    self.fidel_space_orderings = fidel_space_orderings
    super(CPMultiFunctionCaller, self).__init__(funcs, domain, descr,
      argmax=argmax, maxval=maxval, argmin=argmin, minval=minval,
      noise_type=noise_type, noise_scale=noise_scale,
      fidel_space=fidel_space, fidel_cost_func=fidel_cost_func, fidel_to_opt=fidel_to_opt
      )
    self._set_up_point_reconfiguration()

  def _set_up_point_reconfiguration(self):
    """ Sets up reconfiguring points if domain orderings etc. are not None. """
    if self.domain_orderings is not None or self.fidel_space_orderings is not None:
      from .cp_domain_utils import get_raw_point_from_processed_point
    if self.domain_orderings is not None:
      self.get_raw_domain_point_from_processed = \
        lambda pt: get_raw_point_from_processed_point(pt, self.domain,
                     self.domain_orderings.index_ordering,
                     self.domain_orderings.dim_ordering)
    if self.fidel_space_orderings is not None:
      self.get_raw_fidel_from_processed = \
        lambda pt: get_raw_point_from_processed_point(pt, self.fidel_space,
                     self.fidel_space_orderings.index_ordering,
                     self.fidel_space_orderings.dim_ordering)

  def _child_get_candidate_fidels(self, domain_point, filter_by_cost=True,
                                  *args, **kwargs):
    """ Returns candidate fidelities at domain_point for the child class.
        If filter_by_cost is true returns only those for which cost is larger than
        fidel_to_opt.
    """
    num_samples = np.clip(100 * self.fidel_space.get_dim(), 100, 8000)
    return sample_from_cp_domain(self.fidel_space, num_samples,
                                 euclidean_sample_type='latin_hc',
                                 integral_sample_type='latin_hc')

  def get_information_gap(self, fidels):
    """ Returns the information gap to fidel_to_opt. """
    if not self.is_mf():
      raise CalledMultiFidelOnSingleFidelCaller(self)
    return [self.fidel_space.compute_distance(self.fidel_to_opt, fidel) for fidel
            in fidels]


# A wrapper on top of MultiFunctionCaller for single function callers ==================
class FunctionCaller(MultiFunctionCaller):
  """ A function caller on Euclidean spaces. """

  def __init__(self, func, *args, **kwargs):
    """ Constructor. """
    self.func = func
    super(FunctionCaller, self).__init__(func, *args, **kwargs)


class EuclideanFunctionCaller(EuclideanMultiFunctionCaller):
  """ An experiment caller on Euclidean spaces. """

  def __init__(self, func, *args, **kwargs):
    """ Constructor. """
    self.func = func
    super(EuclideanFunctionCaller, self).__init__(func, *args, **kwargs)


class CPFunctionCaller(CPMultiFunctionCaller):
  """ Cartesian product function caller. """

  def __init__(self, func, domain, descr='', raw_func=None, *args, **kwargs):
    """ Constructor. """
    self.func = func
    self.raw_func = raw_func
    super(CPFunctionCaller, self).__init__(func, domain, descr, raw_func, *args, **kwargs)


# An API to obtain a MultiFunctionCaller on a CPDomain from cp_domain ==================
def get_multifunction_caller_from_config(raw_funcs, domain_config_file, descr='',
                                         raw_fidel_cost_func=None, **kwargs):
  """ Returns a multi-function caller from the raw functions and configuration file. """
  # pylint: disable=no-member
  # pylint: disable=maybe-no-member
  if isinstance(domain_config_file, str):
    config = load_config_file(domain_config_file)
  else:
    config = domain_config_file
  if isinstance(raw_funcs, (list, tuple)):
    is_multi_function = True
  else:
    is_multi_function = False
    raw_funcs = [raw_funcs]
  funcs = [get_processed_func_from_raw_func_via_config(rf, config) for rf in raw_funcs]
  # Check multi-fidelity
  if hasattr(config, 'fidel_space') and config.fidel_space is not None:
    fidel_cost_func = get_processed_func_from_raw_func_for_cp_domain(raw_fidel_cost_func,
      config.fidel_space, config.fidel_space_orderings.index_ordering,
      config.fidel_space_orderings.dim_ordering)
    fidel_space_orderings = config.fidel_space_orderings
    fidel_to_opt = config.fidel_to_opt
    fidel_space = config.fidel_space
  else:
    fidel_cost_func = None
    fidel_space_orderings = None
    fidel_to_opt = None
    fidel_space = None
  # Now return
  if is_multi_function:
    return CPMultiFunctionCaller(funcs, config.domain, descr, raw_funcs,
             config.domain_orderings, fidel_space=fidel_space,
             fidel_cost_func=fidel_cost_func, fidel_to_opt=fidel_to_opt,
             fidel_space_orderings=fidel_space_orderings, **kwargs)
  else:
    return CPFunctionCaller(funcs[0], config.domain, descr, raw_funcs[0],
             config.domain_orderings, fidel_space=fidel_space,
             fidel_cost_func=fidel_cost_func, fidel_to_opt=fidel_to_opt,
             fidel_space_orderings=fidel_space_orderings, **kwargs)




# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: config.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='config.proto',
  package='',
  serialized_pb=_b('\n\x0c\x63onfig.proto\"\x81\x02\n\nExperiment\x12\x0c\n\x04name\x18\x01 \x02(\t\x12)\n\x06\x64omain\x18\x02 \x03(\x0b\x32\x19.Experiment.ParameterSpec\x12.\n\x0b\x66idel_space\x18\x03 \x03(\x0b\x32\x19.Experiment.ParameterSpec\x1a\x89\x01\n\rParameterSpec\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\x13\n\x04type\x18\x02 \x01(\t:\x05\x66loat\x12\x11\n\x03min\x18\x03 \x01(\x01:\x04-inf\x12\x10\n\x03max\x18\x04 \x01(\x01:\x03inf\x12\r\n\x03\x64im\x18\x05 \x01(\t:\x00\x12\x10\n\x06kernel\x18\x06 \x01(\t:\x00\x12\x0f\n\x05items\x18\x07 \x01(\t:\x00')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_EXPERIMENT_PARAMETERSPEC = _descriptor.Descriptor(
  name='ParameterSpec',
  full_name='Experiment.ParameterSpec',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Experiment.ParameterSpec.name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='type', full_name='Experiment.ParameterSpec.type', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=_b("float").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='min', full_name='Experiment.ParameterSpec.min', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=True, default_value=-1e10000,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='max', full_name='Experiment.ParameterSpec.max', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=True, default_value=1e10000,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dim', full_name='Experiment.ParameterSpec.dim', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='kernel', full_name='Experiment.ParameterSpec.kernel', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='items', full_name='Experiment.ParameterSpec.items', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=137,
  serialized_end=274,
)

_EXPERIMENT = _descriptor.Descriptor(
  name='Experiment',
  full_name='Experiment',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Experiment.name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='domain', full_name='Experiment.domain', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='fidel_space', full_name='Experiment.fidel_space', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_EXPERIMENT_PARAMETERSPEC, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17,
  serialized_end=274,
)

_EXPERIMENT_PARAMETERSPEC.containing_type = _EXPERIMENT
_EXPERIMENT.fields_by_name['domain'].message_type = _EXPERIMENT_PARAMETERSPEC
_EXPERIMENT.fields_by_name['fidel_space'].message_type = _EXPERIMENT_PARAMETERSPEC
DESCRIPTOR.message_types_by_name['Experiment'] = _EXPERIMENT

Experiment = _reflection.GeneratedProtocolMessageType('Experiment', (_message.Message,), dict(

  ParameterSpec = _reflection.GeneratedProtocolMessageType('ParameterSpec', (_message.Message,), dict(
    DESCRIPTOR = _EXPERIMENT_PARAMETERSPEC,
    __module__ = 'config_pb2'
    # @@protoc_insertion_point(class_scope:Experiment.ParameterSpec)
    ))
  ,
  DESCRIPTOR = _EXPERIMENT,
  __module__ = 'config_pb2'
  # @@protoc_insertion_point(class_scope:Experiment)
  ))
_sym_db.RegisterMessage(Experiment)
_sym_db.RegisterMessage(Experiment.ParameterSpec)


# @@protoc_insertion_point(module_scope)



"""
  Unit tests for continuous distributions.
  -- kvysyara@andrew.cmu.edu
"""
from __future__ import absolute_import

# pylint: disable=invalid-name
# pylint: disable=no-self-use

import numpy as np

# Local imports
from .continuous import Normal, Exponential, ContinuousUniform
from .product import JointDistribution
from ..utils.base_test_class import BaseTestClass, execute_tests

class JointDistributionTestCase(BaseTestClass):
  """ Unit tests for joint distribution class """

  def setUp(self):
    """ Sets up unit tests. """
    self.size = 1000000
    self.threshold = 0.1

  def _check_sample_sizes(self, samples):
    """ Compares the sample sizes with the size parameter"""
    assert self.size == len(samples)

  def _compute_mean(self, samples, axis=None):
    """ Computes Mean """
    return np.mean(samples, axis)

  def _compute_variance(self, samples, axis=None):
    """ Computes Variance """
    return np.var(samples, axis)

  def _compute_covariance(self, samples):
    """ Computes Covariance """
    return np.cov(samples.T)

  def test_rand_sampling(self):
    """ Tests random sampling from Joint distribution """
    self.report('Test random sampling of Joint Distribution.')
    normal_rv = Normal(0, 1)
    exp_rv = Exponential(2)
    uniform_rv = ContinuousUniform(-5, 5)
    dist = JointDistribution([normal_rv, exp_rv, uniform_rv])
    samples = dist.draw_samples('random', self.size)
    for i, rv in enumerate(dist.get_rv_s()):
      self._check_sample_sizes(samples[i])
      assert abs(rv.get_mean() - self._compute_mean(samples[i])) <= self.threshold
      assert abs(rv.get_variance() - self._compute_variance(samples[i])) <= self.threshold

if __name__ == '__main__':
  execute_tests()




"""
  Unit tests for Random CP optimiser on Cartesian product domains.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name
# pylint: disable=abstract-class-little-used


import os
# Local imports
from ..test_data.multiobjective_hartmann.multiobjective_hartmann \
         import objectives as moo_hartmann
from ..test_data.multiobjective_park.multiobjective_park \
         import objectives as moo_park
from ..exd.cp_domain_utils import get_raw_point_from_processed_point, \
                                load_config_file
from ..exd.experiment_caller import get_multifunction_caller_from_config
from ..exd.worker_manager import SyntheticWorkerManager
from . import random_multiobjective_optimiser
from ..utils.base_test_class import BaseTestClass, execute_tests
from ..utils.reporters import get_reporter




class CPMultiObjectiveOptimiserBaseTestCase(object):
  """ Base test class for optimisers on Cartesian product spaces. """
  # pylint: disable=no-member

  def setUp(self):
    """ Set up. """
    self.max_capital = 20
    self._child_set_up()
    self.worker_manager_1 = SyntheticWorkerManager(1, time_distro='const')
    self.worker_manager_3 = SyntheticWorkerManager(3, time_distro='halfnormal')
    file_dir = os.path.dirname(os.path.realpath(__file__))
    test_data_pardir = os.path.dirname(file_dir)
    self.opt_problems = [
      (test_data_pardir + '/test_data/multiobjective_hartmann/config.json',
                                                           (moo_hartmann,)),
      (test_data_pardir + '/test_data/multiobjective_park/config.json', (moo_park,)),
      ]

  def _child_set_up(self):
    """ Child set up. """
    pass

  @classmethod
  def _child_instantiate_optimiser(cls, func_caller, worker_manager, options, reporter):
    """ Instantiate the optimiser. """
    raise NotImplementedError('Implement in a child class.')

  @classmethod
  def _run_optimiser(cls, raw_funcs, domain_config_file, worker_manager, max_capital,
                     mode, *args, **kwargs):
    """ Run the optimiser from given args. """
    raise NotImplementedError('Implement in a child class.')

  def test_instantiation(self):
    """ Tests instantiation of the optimiser. """
    self.report('Test instantiation of multi-objective optimiser.')
    for idx, (dcf, (raw_prob_funcs, )) in enumerate(self.opt_problems):
      self.report('[%d/%d] Testing instantiation of optimiser for %s.'%(
                  idx + 1, len(self.opt_problems), dcf), 'test_result')
      config = load_config_file(dcf)
      multi_func_caller = get_multifunction_caller_from_config(raw_prob_funcs, config)
      optimiser = self._child_instantiate_optimiser(
        multi_func_caller, self.worker_manager_3, options=None,
        reporter=get_reporter('silent'))
      self.report('Instantiated %s object.'%(type(optimiser)))
      for attr in dir(optimiser):
        if not attr.startswith('_'):
          self.report('optimiser.%s = %s'%(attr, str(getattr(optimiser, attr))),
                      'test_result')

  def _test_optimiser_results(self, raw_prob_funcs, pareto_vals, pareto_points,
                              history, dcf):
    """ Tests optimiser results. """
    config = load_config_file(dcf)
    multi_func_caller = get_multifunction_caller_from_config(raw_prob_funcs, config)
    raw_pareto_points = [get_raw_point_from_processed_point(pop, config.domain,
                                         config.domain_orderings.index_ordering,
                                         config.domain_orderings.dim_ordering)
                         for pop in pareto_points]
    self.report('Pareto opt point [-1]: proc=%s, raw=%s.'%(pareto_points[-1],
                                                           raw_pareto_points[-1]))
    saved_in_history = [key for key, _ in list(history.__dict__.items()) if not
                        key.startswith('__')]
    self.report('Stored in history: %s.'%(saved_in_history), 'test_result')
    assert len(history.curr_pareto_vals) == len(history.curr_pareto_points)
    for val in pareto_vals:
      assert len(val) == multi_func_caller.num_funcs
    for pt in pareto_points:
      assert len(pt) == config.domain.num_domains
    self.report('Pareto optimal points: %s.'%(pareto_points))
    self.report('Pareto optimal values: %s.'%(pareto_vals))

  def test_optimisation_single(self):
    """ Test optimisation with a single worker. """
    self.report('')
    self.report('Testing %s with one worker.'%(type(self)))
    for idx, (dcf, (raw_prob_funcs, )) in enumerate(self.opt_problems):
      self.report('[%d/%d] Testing optimisation  with 1 worker on %s.'%(
                  idx + 1, len(self.opt_problems), dcf), 'test_result')
      self.worker_manager_1.reset()
      pareto_vals, pareto_points, history = self._run_optimiser(raw_prob_funcs, dcf,
        self.worker_manager_1, self.max_capital, 'asy')
      self._test_optimiser_results(raw_prob_funcs, pareto_vals, pareto_points, history,
                                   dcf)
      self.report('')

  def test_optimisation_asynchronous(self):
    """ Testing random optimiser with three asynchronous workers. """
    self.report('')
    self.report('Testing %s with three asynchronous workers.'%(type(self)))
    for idx, (dcf, (raw_prob_funcs, )) in enumerate(self.opt_problems):
      self.report('[%d/%d] Testing optimisation  with 3 asynchronous workers on %s.'%(
                  idx + 1, len(self.opt_problems), dcf), 'test_result')
      self.worker_manager_3.reset()
      pareto_vals, pareto_points, history = self._run_optimiser(raw_prob_funcs, dcf,
        self.worker_manager_3, self.max_capital, 'asy')
      self._test_optimiser_results(raw_prob_funcs, pareto_vals, pareto_points, history,
                                   dcf)
      self.report('')


class CPRandomMultiObjectiveOptimiserTestCase(
  CPMultiObjectiveOptimiserBaseTestCase, BaseTestClass):
  """ Unit tests for random multi-objective optimisation. """

  @classmethod
  def _child_instantiate_optimiser(cls, multi_func_caller, worker_manager, options,
                                   reporter):
    """ Instantiate optimiser. """
    return random_multiobjective_optimiser.CPRandomMultiObjectiveOptimiser(
             multi_func_caller, worker_manager, options, reporter)

  @classmethod
  def _run_optimiser(cls, raw_prob_funcs, domain_config_file, worker_manager, max_capital,
                     mode, *args, **kwargs):
    """ Runs multi-objective optimiser. """
    rmoo = random_multiobjective_optimiser
    return rmoo.cp_random_multiobjective_optimisation_from_raw_args(raw_prob_funcs,
             domain_config_file, worker_manager, max_capital, mode, *args, **kwargs)


if __name__ == '__main__':
  execute_tests()




"""
  Unit tests for ed_utils.
  -- kandasamy@cs.cmu.edu
"""
from __future__ import absolute_import
from __future__ import division

# pylint: disable=relative-import

# Local imports
from .exd_utils import random_sampling_cts, random_sampling_kmeans_cts
from ..utils.base_test_class import BaseTestClass, execute_tests


class EDUtilsTestCase(BaseTestClass):
  """ Unit tests for generic functions ed_utils.py """

  def setUp(self):
    """ Sets up unit tests. """
    self.lhs_data = [(1, 10), (2, 5), (4, 10), (10, 100)]

  @classmethod
  def _check_sample_sizes(cls, data, samples):
    """ Data is a tuple of the form (dim, num_samples) ans samples is an ndarray."""
    assert (data[1], data[0]) == samples.shape

  def test_random_sampling(self):
    """ Tests random sampling. """
    self.report('Test random sampling.')
    for data in self.lhs_data:
      self._check_sample_sizes(data, random_sampling_cts(data[0], data[1]))

  def test_random_sampling_kmeans(self):
    """ Tests random sampling with k-means. """
    self.report('Test random sampling with k-means.')
    for data in self.lhs_data:
      self._check_sample_sizes(data, random_sampling_kmeans_cts(data[0], data[1]))


if __name__ == '__main__':
  execute_tests()




"""
  Implements some utilities for Cartesian Product GPs.
  - The term config refers to a namespace that stores information about the optimisation
    problem. Typically, it is what is returened by the load_config_file function.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name

from __future__ import print_function

from argparse import Namespace
from copy import deepcopy
# Local imports
from . import domains
from ..parse.config_parser import config_parser
from ..utils.general_utils import flatten_list_of_objects_and_iterables, \
                                get_original_order_from_reordered_list, \
                                transpose_list_of_lists
from ..utils.oper_utils import random_sample_from_euclidean_domain, \
                               random_sample_from_discrete_euclidean_domain, \
                               random_sample_from_integral_domain, \
                               random_sample_from_prod_discrete_domain


def _process_fidel_to_opt(raw_fidel_to_opt, fidel_space, fidel_space_orderings,
                          config_file):
  """ Processes and returns fidel_to_opt. """
  if raw_fidel_to_opt is None:
    fidel_to_opt = None
    from warnings import warn
    warn('fidel_to_opt is None for %s.'%(config_file))
  else:
    try:
      fidel_to_opt = get_processed_point_from_raw_point(raw_fidel_to_opt,
                       fidel_space, fidel_space_orderings.index_ordering,
                       fidel_space_orderings.dim_ordering)
      assert fidel_space.is_a_member(fidel_to_opt)
    except:
      raise ValueError('fidel_to_opt: %s (raw: %s) is not a member of fidel_space %s.'%(
                       fidel_to_opt, raw_fidel_to_opt, fidel_space))
  # Return
  return raw_fidel_to_opt, fidel_to_opt


def _preprocess_domain_parameters(domain_parameters, var_prefix='var_'):
  """ Preprocesses domain parameters in a configuration specification. """
  for idx, var_dict in enumerate(domain_parameters):
    if not 'name' in var_dict.keys():
      var_dict['name'] = '%s%02d'%(var_prefix, idx)
    if not 'dim' in var_dict.keys():
      var_dict['dim'] = ''
    if not 'kernel' in var_dict.keys():
      var_dict['kernel'] = ''
    if var_dict['type'] in ['float', 'int']:
      if not ('min' in var_dict.keys() and 'max' in var_dict.keys()):
        if not 'bounds' in var_dict.keys():
          raise ValueError('Specify bounds or min and max for Euclidean and Integral ' +
                           'variables: %s.'%(var_dict))
        else:
          var_dict['min'] = var_dict['bounds'][0]
          var_dict['max'] = var_dict['bounds'][1]
    if var_dict['type'] == 'discrete_euclidean' and var_dict['dim'] != '':
      raise ValueError('dim parameter for Discrete Euclidean vectors should be an empty' +
                       ' string or not specified. Given %s.'%(var_dict['dim']))
  return domain_parameters


def _preprocess_config_params(config_params):
  """ Preprocesses configuration parameters. """
  config_params = deepcopy(config_params)
  # The name of the experiment
  if not 'name' in config_params:
    if 'exp_info' in config_params and 'name' in config_params['exp_info']:
      config_params['name'] = config_params['exp_info']['name']
    else:
      config_params['name'] = 'no_name'
  # Process the domain variables
  config_params['domain'] = _preprocess_domain_parameters(config_params['domain'],
                              var_prefix='domvar_')
  if 'fidel_space' in config_params:
    config_params['fidel_space'] = _preprocess_domain_parameters(
                                     config_params['fidel_space'], var_prefix='fidelvar_')
  return config_params


def load_config_file(config_file, *args, **kwargs):
  """ Loads the configuration file. """
  parsed_result = config_parser(config_file)
  # If loading from file, no need to pre-process
  return load_config(parsed_result, config_file, *args, **kwargs)


def load_config(config_params, config_file=None, *args, **kwargs):
  """ Loads configuration from parameters. """
  config_params = _preprocess_config_params(config_params)
  domain_params = config_params['domain']
  domain_constraints = None if not ('domain_constraints' in config_params.keys()) \
                       else config_params['domain_constraints']
  domain_info = Namespace(config_file=config_file)
  domain, domain_orderings = load_domain_from_params(domain_params,
    domain_constraints=domain_constraints, domain_info=domain_info, *args, **kwargs)
  config = Namespace(name=config_params['name'],
                     domain=domain, domain_orderings=domain_orderings)
  # Check the fidelity space
  if 'fidel_space' in config_params.keys():
    fidel_space_params = config_params['fidel_space']
    fidel_space_constraints = None if not ('fidel_space_constraints' in
                                           config_params.keys()) \
                              else config_params['fidel_space_constraints']
    fidel_space_info = Namespace(config_file=config_file)
    fidel_space, fidel_space_orderings = load_domain_from_params(
      fidel_space_params, domain_constraints=fidel_space_constraints,
      domain_info=fidel_space_info, *args, **kwargs)
    if len(fidel_space.list_of_domains) > 0:
      config.fidel_space = fidel_space
      config.fidel_space_orderings = fidel_space_orderings
      config.raw_fidel_to_opt, config.fidel_to_opt = _process_fidel_to_opt(
        config_params['fidel_to_opt'], fidel_space, fidel_space_orderings,
        config_file)
  return config


def load_cp_domain_from_config_file(config_file, *args, **kwargs):
  """ Loads and creates a cartesian product domain object from a config_file. """
  parsed_result = config_parser(config_file)
  domain_params = parsed_result['domain']
  domain_constraints = None if not ('domain_constraints' in parsed_result.keys()) \
                       else parsed_result['domain_constraints']
  domain_info = Namespace(config_file=config_file)
  return load_domain_from_params(domain_params, domain_constraints=domain_constraints,
                                 domain_info=domain_info, *args, **kwargs)


def load_domain_from_params(domain_params,
      general_euclidean_kernel='', general_integral_kernel='',
      general_discrete_kernel='', general_discrete_numeric_kernel='',
      domain_constraints=None, domain_info=None):
  """ Loads and creates a cartesian product object from a config_file. """
  # pylint: disable=too-many-branches
  # pylint: disable=too-many-statements
  # pylint: disable=too-many-locals
  list_of_domains = []
  general_euclidean_bounds = []
  general_euclidean_idxs = []
  general_integral_bounds = []
  general_integral_idxs = []
  general_discrete_items_list = []
  general_discrete_idxs = []
  general_discrete_numeric_items_list = []
  general_discrete_numeric_idxs = []
  raw_name_ordering = []
  # We will need the following variables for the function caller and the kernel
  index_ordering = [] # keeps track of which index goes where in the domain
  # iterate over the loop
  for idx, param in enumerate(domain_params):
    raw_name_ordering.append(param['name'])
    if param['type'] in ['float', 'int']:
      bound_dim = 1 if param['dim'] == '' else param['dim']
      curr_bounds = [[param['min'], param['max']]] * bound_dim
    elif param['type'] in ['discrete', 'discrete_numeric', 'boolean',
                           'discrete_euclidean']:
      items_dim = 1 if param['dim'] == '' else param['dim']
      if param['type'] == 'boolean':
        param_items = [0, 1]
      else:
        param_items = param['items']
      curr_items = [param_items[:] for _ in range(items_dim)]
    # Now append to relevant part
    if param['type'] == 'float':
      if param['kernel'] == '':
        general_euclidean_bounds.extend(curr_bounds)
        general_euclidean_idxs.append(idx)
      else:
        list_of_domains.append(domains.EuclideanDomain(curr_bounds))
        index_ordering.append(idx)
    elif param['type'] == 'int':
      if param['kernel'] == '':
        general_integral_bounds.extend(curr_bounds)
        general_integral_idxs.append(idx)
      else:
        list_of_domains.append(domains.IntegralDomain(curr_bounds))
        index_ordering.append(idx)
    elif param['type'] in ['boolean', 'discrete']:
      if param['kernel'] == '':
        general_discrete_items_list.extend(curr_items)
        general_discrete_idxs.append(idx)
      else:
        list_of_domains.append(domains.ProdDiscreteDomain(curr_items))
        index_ordering.append(idx)
    elif param['type'] == 'discrete_numeric':
      if param['kernel'] == '':
        general_discrete_numeric_items_list.extend(curr_items)
        general_discrete_numeric_idxs.append(idx)
      else:
        list_of_domains.append(domains.ProdDiscreteNumericDomain(curr_items))
        index_ordering.append(idx)
    elif param['type'] == 'discrete_euclidean':
      # We will treat the discrete Euclidean space differently
      list_of_domains.append(domains.DiscreteEuclideanDomain(param_items))
      index_ordering.append(idx)
    elif param['type'].startswith(('nn', 'cnn', 'mlp')):
      from ..nn.nn_domains import get_nn_domain_from_constraints
      list_of_domains.append(get_nn_domain_from_constraints(param['type'], param))
      index_ordering.append(idx)
    else:
      raise ValueError('Unknown domain type: %s.'%(param['type']))
  # Create kernel ordering and variable ordering
  kernel_ordering = [domain_params[idx]['kernel'] for idx in index_ordering]
  name_ordering = [domain_params[idx]['name'] for idx in index_ordering]
  dim_ordering = [domain_params[idx]['dim'] for idx in index_ordering]
  # For general euclidean and integral domains
  if len(general_euclidean_bounds) > 0:
    list_of_domains.append(domains.EuclideanDomain(general_euclidean_bounds))
    general_euclidean_names = [domain_params[idx]['name'] for idx in
                               general_euclidean_idxs]
    general_euclidean_dims = [domain_params[idx]['dim'] for idx in general_euclidean_idxs]
    name_ordering.append(general_euclidean_names)
    dim_ordering.append(general_euclidean_dims)
    index_ordering.append(general_euclidean_idxs)
    kernel_ordering.append(general_euclidean_kernel)
  if len(general_integral_bounds) > 0:
    list_of_domains.append(domains.IntegralDomain(general_integral_bounds))
    general_integral_names = [domain_params[idx]['name'] for idx in general_integral_idxs]
    general_integral_dims = [domain_params[idx]['dim'] for idx in general_integral_idxs]
    name_ordering.append(general_integral_names)
    dim_ordering.append(general_integral_dims)
    index_ordering.append(general_integral_idxs)
    kernel_ordering.append(general_integral_kernel)
  if len(general_discrete_items_list) > 0:
    list_of_domains.append(domains.ProdDiscreteDomain(general_discrete_items_list))
    general_discrete_names = [domain_params[idx]['name'] for idx in general_discrete_idxs]
    general_discrete_dims = [domain_params[idx]['dim'] for idx in general_discrete_idxs]
    name_ordering.append(general_discrete_names)
    dim_ordering.append(general_discrete_dims)
    index_ordering.append(general_discrete_idxs)
    kernel_ordering.append(general_discrete_kernel)
  if len(general_discrete_numeric_items_list) > 0:
    list_of_domains.append(
      domains.ProdDiscreteNumericDomain(general_discrete_numeric_items_list))
    general_discrete_numeric_names = \
      [domain_params[idx]['name'] for idx in general_discrete_numeric_idxs]
    general_discrete_numeric_dims = \
      [domain_params[idx]['dim'] for idx in general_discrete_numeric_idxs]
    name_ordering.append(general_discrete_numeric_names)
    dim_ordering.append(general_discrete_numeric_dims)
    index_ordering.append(general_discrete_numeric_idxs)
    kernel_ordering.append(general_discrete_numeric_kernel)
  # Arrange all orderings into a namespace
  orderings = Namespace(index_ordering=index_ordering,
                        kernel_ordering=kernel_ordering,
                        dim_ordering=dim_ordering,
                        name_ordering=name_ordering,
                        raw_name_ordering=raw_name_ordering)
  # Create a namespace with additional information
  if domain_info is None:
    domain_info = Namespace()
  domain_info.config_orderings = orderings
  if domain_constraints is not None:
    domain_info.constraints = domain_constraints
  # Create a cartesian product domain
  cp_domain = domains.CartesianProductDomain(list_of_domains, domain_info)
  return cp_domain, orderings


def get_num_raw_domains(ordering):
  """ Returns the number of raw domains. """
  num_raw_domains = len(ordering)
  for elem in ordering:
    if hasattr(elem, '__iter__'):
      num_raw_domains += len(elem) - 1
  return num_raw_domains


def _unpack_vectorised_domain(x, dim_ordering):
  """ Unpacks a vectorised domain. """
  ret = [None] * len(dim_ordering)
  counter = 0
  for idx, num_dims in enumerate(dim_ordering):
    if num_dims == '':
      ret[idx] = x[counter]
      counter += 1
    else:
      ret[idx] = x[counter:counter+num_dims]
      counter += num_dims
  assert counter == len(x) # Check if number of variables match
  return ret


# Functions to pack and unpack points in the domain -----------------------------------
def get_processed_point_from_raw_point(raw_x, cp_domain, index_ordering, dim_ordering):
  """ Obtains the processed point from the raw point. """
  if not cp_domain.get_type() == 'cartesian_product':
    packed_x = [raw_x[index_ordering[j]] for j in index_ordering]
    return flatten_list_of_objects_and_iterables(packed_x)
  else:
    packed_x = [None] * len(index_ordering)
    for idx, idx_order in enumerate(index_ordering):
      if isinstance(idx_order, list):
        curr_elem = [raw_x[j] for j in idx_order]
        curr_elem = flatten_list_of_objects_and_iterables(curr_elem)
        packed_x[idx] = curr_elem
      elif dim_ordering[idx] == '' and (cp_domain.list_of_domains[idx].get_type() in \
                     ['euclidean', 'integral', 'prod_discrete', 'prod_discrete_numeric']):
        packed_x[idx] = [raw_x[idx_order]]
      else:
        packed_x[idx] = raw_x[idx_order]
    return packed_x


def get_raw_point_from_processed_point(proc_x, cp_domain, index_ordering, dim_ordering):
  """ Gets the raw point from the processed point. """
  if not cp_domain.get_type() == 'cartesian_product':
    repacked_x = _unpack_vectorised_domain(proc_x, dim_ordering)
  else:
    repacked_x = []
    for idx, raw_dim in enumerate(dim_ordering):
      if cp_domain.list_of_domains[idx].get_type() == 'discrete_euclidean':
        repacked_x.append([proc_x[idx]])
      elif isinstance(raw_dim, list):
        repacked_x.append(_unpack_vectorised_domain(proc_x[idx], raw_dim))
      elif raw_dim == '':
        repacked_x.append(proc_x[idx])
      else:
        repacked_x.append([proc_x[idx]])
    repacked_x = flatten_list_of_objects_and_iterables(repacked_x)
  flattened_index_ordering = flatten_list_of_objects_and_iterables(index_ordering)
  x_orig_order = get_original_order_from_reordered_list(repacked_x,
                                                        flattened_index_ordering)
  return x_orig_order


def get_raw_from_processed_via_config(proc_point, config):
  """ Gets the raw point (both domain and fidel) from the processed point. """
  has_fidel = hasattr(config, 'fidel_space')
  if has_fidel:
    proc_fidelity, proc_dom_point = proc_point
  else:
    proc_dom_point = proc_point
  raw_dom_point = get_raw_point_from_processed_point(proc_dom_point, config.domain,
    config.domain_orderings.index_ordering, config.domain_orderings.dim_ordering)
  if has_fidel:
    raw_fidelity = get_raw_point_from_processed_point(proc_fidelity, config.fidel_space,
                                              config.fidel_space_orderings.index_ordering,
                                              config.fidel_space_orderings.dim_ordering)
    return [raw_fidelity, raw_dom_point]
  else:
    return raw_dom_point


def get_processed_from_raw_via_config(raw_point, config):
  """ Gets the processed point (both domain and fidel) from the raw point. """
  has_fidel = hasattr(config, 'fidel_space')
  if has_fidel:
    raw_fidelity, raw_dom_point = raw_point
  else:
    raw_dom_point = raw_point
  proc_dom_point = get_processed_point_from_raw_point(raw_dom_point, config.domain,
    config.domain_orderings.index_ordering, config.domain_orderings.dim_ordering)
  if has_fidel:
    proc_fidelity = get_processed_point_from_raw_point(raw_fidelity, config.fidel_space,
                      config.fidel_space_orderings.index_ordering,
                      config.fidel_space_orderings.dim_ordering)
    return [proc_fidelity, proc_dom_point]
  else:
    return proc_dom_point


# Functions to sample from cp_domain ----------------------------------------------------
def sample_from_cp_domain(cp_domain, num_samples, domain_samplers=None,
                          euclidean_sample_type='rand',
                          integral_sample_type='rand',
                          nn_sample_type='rand',
                          discrete_euclidean_sample_type='rand',
                          max_num_retries_for_constraint_satisfaction=10,
                          verbose_constraint_satisfaction=True):
  """ Samples from the CP domain. """
  ret = []
  if cp_domain.has_constraints():
    num_samples_to_draw = max(10, 2 * num_samples)
  else:
    num_samples_to_draw = num_samples
  for _ in range(max_num_retries_for_constraint_satisfaction):
    curr_ret = sample_from_cp_domain_without_constraints(cp_domain, num_samples_to_draw,
                 domain_samplers, euclidean_sample_type, integral_sample_type,
                 nn_sample_type, discrete_euclidean_sample_type)
    # Check constraints
    if cp_domain.has_constraints():
      constraint_satisfying_ret = [elem for elem in curr_ret if
                                   cp_domain.constraints_are_satisfied(elem)]
      curr_ret = constraint_satisfying_ret
    ret.extend(curr_ret)
    # Check length of ret
    if len(ret) >= num_samples:
      ret = ret[:num_samples]
      break
  # Check if the number of samples is too small and print a warning accordingly.
  if len(ret) < num_samples and verbose_constraint_satisfaction:
    if len(ret) == 0:
      Warning(('sample_from_cp_domain obtained 0 samples (%d requested) despite %d ' +
       'tries. This is most likely because your constraints specify a set of measure 0.' +
       'Consider parametrising your domain differently.')%(
       num_samples, max_num_retries_for_constraint_satisfaction))
    elif len(ret)/float(num_samples) < 0.25:
      Warning(('sample_from_cp_domain obtained %d samples (%d requested) despite %d ' +
       'tries. This is because your constraints specify a very small subset of the '
       'original domain. Consider parametrising your domain differently.')%(
       len(ret), num_samples, max_num_retries_for_constraint_satisfaction))
    else:
      Warning(('sample_from_cp_domain was only able to obtain %d samples (%d requested)' +
       ' despite %d tries. Try increasing max_num_retries_for_constraint_satisfaction.')%(
       len(ret), num_samples, max_num_retries_for_constraint_satisfaction))
  return ret


def sample_from_cp_domain_without_constraints(cp_domain, num_samples,
                                              domain_samplers=None,
                                              euclidean_sample_type='rand',
                                              integral_sample_type='rand',
                                              nn_sample_type='rand',
                                              discrete_euclidean_sample_type='rand'):
  """ Samples from the CP domain without the constraints. """
  if domain_samplers is None:
    domain_samplers = [None] * cp_domain.num_domains
  individual_domain_samples = []
  for idx, dom in enumerate(cp_domain.list_of_domains):
    if domain_samplers[idx] is not None:
      curr_domain_samples = domain_samplers[idx](num_samples)
    else:
      if dom.get_type() == 'euclidean':
        curr_domain_samples = random_sample_from_euclidean_domain(dom.bounds, num_samples,
                                                                  euclidean_sample_type)
      elif dom.get_type() == 'discrete_euclidean':
        curr_domain_samples = random_sample_from_discrete_euclidean_domain(
                                dom.list_of_items, num_samples,
                                discrete_euclidean_sample_type)
      elif dom.get_type() == 'integral':
        curr_domain_samples = random_sample_from_integral_domain(dom.bounds, num_samples,
                                                                 integral_sample_type)
      elif dom.get_type() in ['prod_discrete', 'prod_discrete_numeric']:
        curr_domain_samples = random_sample_from_prod_discrete_domain(
                                dom.list_of_list_of_items, num_samples)
      elif dom.get_type() == 'neural_network':
        from ..nn.nn_opt_utils import random_sample_from_nn_domain
        curr_domain_samples = random_sample_from_nn_domain(dom.nn_type, num_samples,
                                                           nn_sample_type,
                                                           dom.constraint_checker)
      elif dom.get_type() == 'cartesian_product':
        curr_domain_samples = sample_from_cp_domain(dom, num_samples,
                    euclidean_sample_type=euclidean_sample_type,
                    integral_sample_type=integral_sample_type,
                    nn_sample_type=nn_sample_type,
                    discrete_euclidean_sample_type=discrete_euclidean_sample_type)
      else:
        raise ValueError('Unknown domain type %s. Provide sampler.'%(dom.get_type()))
      individual_domain_samples.append(curr_domain_samples)
  return transpose_list_of_lists(individual_domain_samples)


def sample_from_config_space(config, num_samples,
                             fidel_space_samplers=None,
                             domain_samplers=None,
                             fidel_space_euclidean_sample_type='rand',
                             fidel_space_integral_sample_type='rand',
                             fidel_space_discrete_euclidean_sample_type='rand',
                             domain_euclidean_sample_type='rand',
                             domain_integral_sample_type='rand',
                             domain_nn_sample_type='rand',
                             domain_discrete_euclidean_sample_type='rand',
                             ):
  """ Samples from the Domain and possibly the fidelity space. """
  # pylint: disable=too-many-arguments
  domain_samples = sample_from_cp_domain(config.domain, num_samples, domain_samplers,
    domain_euclidean_sample_type, domain_integral_sample_type, domain_nn_sample_type,
    domain_discrete_euclidean_sample_type)
  if hasattr(config, 'fidel_space'):
    fidel_space_samples = sample_from_cp_domain(config.fidel_space, num_samples,
      fidel_space_samplers, fidel_space_euclidean_sample_type,
      fidel_space_integral_sample_type, fidel_space_discrete_euclidean_sample_type)
    return [list(x) for x in zip(fidel_space_samples, domain_samples)]
  else:
    return domain_samples


# Functions to get processed raw function -----------------------------------------------
def get_processed_func_from_raw_func_for_cp_domain(raw_func, cp_domain,
                                                   index_ordering, dim_ordering):
  """ Returns a processed function from the raw function based on the ordering.
  """
  # This function does the evaluation.
  def _eval_func_x_after_unpacking(x, _raw_func, _cp_domain, _index_ordering,
                                   _dim_ordering):
    """ Evaluates the function after processing it. """
    return _raw_func(get_raw_point_from_processed_point(x, _cp_domain, _index_ordering,
                                                        _dim_ordering))
  # This function returns the raw function
  def _get_processed_func(_raw_func, _cp_domain, _index_ordering, _dim_ordering):
    """ Returns a function which evaluates raw_func from a packed input. """
    return lambda x: _eval_func_x_after_unpacking(x, _raw_func, _cp_domain,
                                                  _index_ordering, _dim_ordering)
  # Return
  return _get_processed_func(raw_func, cp_domain, index_ordering, dim_ordering)


def get_processed_func_from_raw_func_for_cp_domain_fidelity(raw_func, config):
  """ Returns a processed function from the raw function based on the ordering.
  """
  # This function does the evaluation.
  def _eval_func_z_x_after_unpacking(z, x, _raw_func, _fidel_space, _domain,
    _fidel_space_index_ordering, _fidel_space_dim_ordering,
    _domain_index_ordering, _domain_dim_ordering):
    """ Evaluates the function after processing it. """
    raw_z = get_raw_point_from_processed_point(z, _fidel_space,
              _fidel_space_index_ordering, _fidel_space_dim_ordering)
    raw_x = get_raw_point_from_processed_point(x, _domain,
              _domain_index_ordering, _domain_dim_ordering)
    return _raw_func(raw_z, raw_x)
  # This returns the raw function
  def _get_processed_func(_raw_func, _fidel_space, _domain,
    _fidel_space_index_ordering, _fidel_space_dim_ordering,
    _domain_index_ordering, _domain_dim_ordering):
    """ Evaluates the function after processing it. """
    return lambda z, x: _eval_func_z_x_after_unpacking(z, x, _raw_func,
      _fidel_space, _domain, _fidel_space_index_ordering, _fidel_space_dim_ordering,
      _domain_index_ordering, _domain_dim_ordering)
  # Return
  return _get_processed_func(raw_func, config.fidel_space, config.domain,
                             config.fidel_space_orderings.index_ordering,
                             config.fidel_space_orderings.dim_ordering,
                             config.domain_orderings.index_ordering,
                             config.domain_orderings.dim_ordering)


def get_processed_func_from_raw_func_via_config(raw_func, config):
  """ Returns a processed function from the raw function based on the ordering.
  """
  if hasattr(config, 'fidel_space'):
    return get_processed_func_from_raw_func_for_cp_domain_fidelity(raw_func, config)
  else:
    return get_processed_func_from_raw_func_for_cp_domain(raw_func, config.domain,
             config.domain_orderings.index_ordering, config.domain_orderings.dim_ordering)




"""
  This module implements OTMANN (Optimal Transport Metrics for Neural Network
  Architectures) as described in the following paper.
  Kandasamy, Neiswanger, Schnider, Poczos, Xing, 2018. "Neural Architecture Search with
  Bayesian Optimisation and Optimal Transport".
  --kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name

import numpy as np
# Local imports
from . import neural_network
from ..gp.kernel import ExpSumOfDistsKernel, SumOfExpSumOfDistsKernel
from ..utils.oper_utils import opt_transport

DFLT_CONN_COST_FUNC = 'linear'
DFLT_KERN_DIST_POWERS = 1
REPLACE_COST_INF_WITH = 7.65432e5
CONV_RES_RAW_COST_FRAC = 0.9

CNN_STRUCTURAL_PENALTY_GROUPS = ['all', 'conv', 'pool', 'fc']
MLP_STRUCTURAL_PENALTY_GROUPS = ['all', 'rectifier', 'sigmoid']
PATH_LENGTH_TYPES = ['shortest', 'longest', 'rw']


def _get_conv_filter_size_cost(labi, labj, conv_scale):
  """ Returns the cost for comparing two different convolutional filters. """
  conv_diff = float(abs(int(labi[-1]) - int(labj[-1])))
  return conv_scale * np.sqrt(conv_diff)


def get_cnn_layer_label_mismatch_penalties(non_assignment_penalty, max_conv_size=7,
                                           conv_scale=None):
  """ Gets the label mismatch matrix for a CNN. """
  conv_scale = np.sqrt(2)/10.0 if conv_scale is None else conv_scale
  cnn_layer_labels = neural_network.get_cnn_layer_labels(max_conv_size)
  num_labels = len(cnn_layer_labels)
  label_penalties = np.zeros((num_labels, num_labels))
  for i in range(num_labels):
    for j in range(i, num_labels):
      labi = cnn_layer_labels[i]
      labj = cnn_layer_labels[j]
      if labi == labj:
        cost = 0.0
      elif (labi.startswith('conv') and labj.startswith('conv')) or \
           (labi.startswith('res') and labj.startswith('res')):
        cost = _get_conv_filter_size_cost(labi, labj, conv_scale)
      elif (labi.startswith('conv') and labj.startswith('res')) or \
           (labi.startswith('res') and labj.startswith('conv')):
        raw_cost = _get_conv_filter_size_cost(labi, labj, conv_scale)
        cost = raw_cost if raw_cost > non_assignment_penalty else \
               (CONV_RES_RAW_COST_FRAC * raw_cost +
                (1-CONV_RES_RAW_COST_FRAC) * non_assignment_penalty)
        # When mapping the a convolutional block to a residual block, set the cost
        # to be in-between the cost for a conv-conv layer and the non_assignment_penalty.
      elif labi.endswith('pool') and labj.endswith('pool'):
        cost = 0.5
      else:
        cost = np.inf
      label_penalties[i, j] = cost * non_assignment_penalty
      label_penalties[j, i] = cost * non_assignment_penalty
  return cnn_layer_labels, label_penalties


def get_mlp_layer_label_mismatch_penalties(non_assignment_penalty, class_or_reg,
                                           list_of_activations=None):
  """ Gets the label mismatch penalty for an MLP. """
  rectifiers = neural_network.MLP_RECTIFIERS
  sigmoids = neural_network.MLP_SIGMOIDS
  non_linear_activations = rectifiers + sigmoids
  mlp_layer_labels = neural_network.get_mlp_layer_labels(class_or_reg,
                                                         list_of_activations)
  num_labels = len(mlp_layer_labels)
  label_penalties = np.zeros((num_labels, num_labels))
  for i in range(num_labels):
    for j in range(i, num_labels):
      labi = mlp_layer_labels[i]
      labj = mlp_layer_labels[j]
      if labi == labj:
        cost = 0.0
      elif (labi in rectifiers and labj in rectifiers) or \
           (labi in sigmoids and labj in sigmoids):
        cost = 0.1
      elif labi in non_linear_activations and labj in non_linear_activations:
        cost = 0.25
      else:
        cost = np.inf
      label_penalties[i, j] = cost * non_assignment_penalty
      label_penalties[j, i] = cost * non_assignment_penalty
  return mlp_layer_labels, label_penalties


class NNDistanceComputer(object):
  """ An abstract class for computing distances between neural networks. """

  def __init__(self):
    """ Constructor. """
    super(NNDistanceComputer, self).__init__()

  def __call__(self, X1, X2, *args, **kwargs):
    """ Evaluates the distances by calling evaluate. """
    return self.evaluate(X1, X2, *args, **kwargs)

  def evaluate(self, X1, X2, *args, **kwargs):
    """ Evaluates the distances between X1 and X2 and returns an n1 x n2 distance matrix.
        If X1 and X2 are single neural networks, returns a scalar. """
    if isinstance(X1, neural_network.NeuralNetwork) and \
       isinstance(X2, neural_network.NeuralNetwork):
      return self.evaluate_single(X1, X2, *args, **kwargs)
    else:
      n1 = len(X1)
      n2 = len(X2)
      X2 = X2 if X2 is not None else X1
      x1_is_x2 = X1 is X2

      all_ret = None
      es_is_iterable = None
      for i, x1 in enumerate(X1):
        X2_idxs = range(i, n2) if x1_is_x2 else range(n2)
        for j in X2_idxs:
          x2 = X2[j]
          # Compute the distances
          curr_ret = self.evaluate_single(x1, x2, *args, **kwargs)
          all_ret, es_is_iterable = self._add_to_all_ret(curr_ret, i, j, n1, n2,
                                                         all_ret, es_is_iterable)
          # Check if we need to do j and i as well.
          if x1_is_x2:
            all_ret, es_is_iterable = self._add_to_all_ret(curr_ret, j, i, n1, n2,
                                                           all_ret, es_is_iterable)
#       if all_ret is not None:
#         print "Computed %d distances between %d and %d (%d) networks."%(
#                 len(all_ret), len(X1), len(X2), X1 is X2)
      return all_ret

  @classmethod
  def _add_to_all_ret(cls, curr_ret, i, j, n1, n2, all_ret=None, es_is_iterable=None):
    """ Adds the current result to all results. """
    if all_ret is None:
      if hasattr(curr_ret, '__iter__'):
        es_is_iterable = True
        all_ret = [np.zeros((n1, n2)) for _ in range(len(curr_ret))]
      else:
        es_is_iterable = False
        all_ret = np.zeros((n1, n2))
    if es_is_iterable:
      for k in range(len(curr_ret)):
        all_ret[k][i, j] = curr_ret[k]
    else:
      all_ret[i, j] = curr_ret
    return all_ret, es_is_iterable

  def evaluate_single(self, x1, x2, *args, **kwargs):
    """ Evaluates the distance between the two networks x1 and x2. """
    raise NotImplementedError('Implement in a child class.')


class OTMANNDistanceComputer(NNDistanceComputer):
  """ A distance between neural networks based on solving a transportation problem. """
  #pylint: disable=attribute-defined-outside-init

  def __init__(self, all_layer_labels, label_mismatch_penalty,
               non_assignment_penalty, structural_penalty_groups, path_length_types,
               dflt_mislabel_coeffs=None,
               dflt_struct_coeffs=None,
               dflt_dist_type=None,
               connectivity_diff_cost_function=DFLT_CONN_COST_FUNC):
    """ Constructor. """
    super(OTMANNDistanceComputer, self).__init__()
    self.all_layer_labels = all_layer_labels
    self.label_mismatch_penalty = label_mismatch_penalty
    self.non_assignment_penalty = non_assignment_penalty
    self.structural_penalty_groups = structural_penalty_groups
    self.path_length_types = path_length_types
    self.all_path_length_categories = [x + '-' + y for x in self.structural_penalty_groups
                                       for y in self.path_length_types]
    self.dflt_mislabel_coeffs = dflt_mislabel_coeffs
    self.dflt_struct_coeffs = dflt_struct_coeffs
    self.dflt_dist_type = dflt_dist_type
    self._set_up_connectivity_cost_function(connectivity_diff_cost_function)

  def _set_up_connectivity_cost_function(self, connectivity_diff_cost_function):
    """ Sets up the connectivity difference cost function. """
    self.connectivity_diff_cost_function = connectivity_diff_cost_function
    if connectivity_diff_cost_function == 'linear':
      self._conn_diff_cost_func = np.abs
    elif connectivity_diff_cost_function == 'sqrt':
      self._conn_diff_cost_func = lambda x: np.sqrt(np.abs(x))
    elif connectivity_diff_cost_function == 'log':
      self._conn_diff_cost_func = lambda x: np.log(np.abs(x))
    elif connectivity_diff_cost_function.startswith('poly'):
      poly_order = float(connectivity_diff_cost_function[4:])
      self._conn_diff_cost_func = lambda x: np.abs(x)**poly_order
    return

  def get_mislabel_cost_matrix(self, x1, x2):
    """ Returns the layer cost matrix from the graphs and label_mismatch_penalty """
    label_idxs_1 = [self.all_layer_labels.index(elem) for elem in x1.layer_labels]
    label_idxs_2 = [self.all_layer_labels.index(elem) for elem in x2.layer_labels]
    return self.label_mismatch_penalty[np.ix_(label_idxs_1, label_idxs_2)]

  def _get_cost_matrix_for_fwd_or_bkwd(self, x1_dists, x2_dists):
    """ Gets the cost matrix for one set of distances. """
    curr_cost_accumulation = np.zeros((x1_dists.shape[0], x2_dists.shape[0]))
    for dim in range(x1_dists.shape[1]):
      curr_diffs = x1_dists[:, dim][:, np.newaxis] - x2_dists[:, dim]
      curr_dim_costs = self._conn_diff_cost_func(curr_diffs)
      curr_cost_accumulation += curr_dim_costs
    curr_cost_matrix = curr_cost_accumulation/float(x1_dists.shape[1])
    return curr_cost_matrix

  def get_struct_cost_matrix(self, x1, x2):
    """ Gets a connectivity cost matrix. """
    x1_bkwd_ip_dists, x1_fwd_op_dists = x1.get_bkwd_ip_fwd_op_dists_of_all_layers(
                                          self.all_path_length_categories)
    x2_bkwd_ip_dists, x2_fwd_op_dists = x2.get_bkwd_ip_fwd_op_dists_of_all_layers(
                                          self.all_path_length_categories)
    bkwd_costs = self._get_cost_matrix_for_fwd_or_bkwd(x1_bkwd_ip_dists, x2_bkwd_ip_dists)
    fwd_costs = self._get_cost_matrix_for_fwd_or_bkwd(x1_fwd_op_dists, x2_fwd_op_dists)
    return (bkwd_costs + fwd_costs)/2

  @classmethod
  def get_ot_cost_matrix(cls, mislabel_cost_matrix, struct_cost_matrix,
                              mislabel_coeff, struct_coeff, non_assignment_penalty,
                              replace_cost_inf_with=REPLACE_COST_INF_WITH):
    """ Adds the two matrices and adds an additional dummy layer at he end of the rows and
        columns. Also makes some synthetic changes to enable OT computation. """
    # Add dummy layer
    cost_matrix = mislabel_coeff * mislabel_cost_matrix + \
                  struct_coeff * struct_cost_matrix
    row_add = non_assignment_penalty * np.ones((1, cost_matrix.shape[1]))
    col_add = non_assignment_penalty * np.ones((cost_matrix.shape[0] + 1, 1))
    col_add[-1] = 0.0
    cost_matrix = np.vstack((cost_matrix, row_add))
    cost_matrix = np.hstack((cost_matrix, col_add))
    # Replace infinities with a large value.
    if replace_cost_inf_with is not None and np.isfinite(replace_cost_inf_with):
      cost_matrix[np.logical_not(np.isfinite(cost_matrix))] = replace_cost_inf_with
    return cost_matrix

  def evaluate_single(self, x1, x2, mislabel_coeffs=None, struct_coeffs=None,
                      dist_type=None):
    """ Evaluates the distances between two networks x1 and x2. dist_type is a
        string with options 'lp' or 'emd'.
    """
    # pylint: disable=arguments-differ
    #  Preprocessing
    mislabel_coeffs = mislabel_coeffs if mislabel_coeffs is not None else \
                      self.dflt_mislabel_coeffs
    struct_coeffs = struct_coeffs if struct_coeffs is not None else \
                    self.dflt_struct_coeffs
    dist_type = dist_type if dist_type is not None else \
                self.dflt_dist_type
    if not hasattr(mislabel_coeffs, '__len__'):
      mislabel_coeffs = [mislabel_coeffs]
    if not hasattr(struct_coeffs, '__len__'):
      struct_coeffs = [struct_coeffs]
    assert len(mislabel_coeffs) == len(struct_coeffs)
    # Compute the types of distances we need to compute.
    types_of_distances = dist_type.split('-')
    # Create data for the transportation problem.
    total_wt_1 = sum(x1.layer_masses)
    total_wt_2 = sum(x2.layer_masses)
    supplies = np.append(x1.layer_masses, total_wt_2)
    demands = np.append(x2.layer_masses, total_wt_1)
    # Get the mislabel and structural cost matrices
    mislabel_cost_matrix = self.get_mislabel_cost_matrix(x1, x2)
    struct_cost_matrix = self.get_struct_cost_matrix(x1, x2)
    # Go through each coefficient and repeat
    ret = []
    for coeff_idx in range(len(mislabel_coeffs)):
      curr_ot_cost_matrix = self.get_ot_cost_matrix(
                              mislabel_cost_matrix, struct_cost_matrix,
                              mislabel_coeffs[coeff_idx], struct_coeffs[coeff_idx],
                              self.non_assignment_penalty)
      _, min_val, emd = opt_transport(supplies, demands, curr_ot_cost_matrix)
      # Below, emd and lp-norm-by-max (or min, sum) are not distances.
      for dt in types_of_distances:
        if dt == 'lp':
          ret.append(min_val)
        elif dt == 'emd':
          ret.append(emd)
        elif dt == 'lp_norm_by_max':
          dist_lp_norm = min_val / max(total_wt_1, total_wt_2)
          ret.append(dist_lp_norm)
        elif dt == 'log_lp':
          ret.append(np.log(1 + min_val))
        else:
          raise ValueError('Unknown dist_type \'%s\'.'%(dist_type))
    return ret


class DistProdNNKernel(ExpSumOfDistsKernel):
  """ Computes a kernel using the transportation distance as the distance. """

  def __init__(self, trans_dist_computer, betas, scale, powers=DFLT_KERN_DIST_POWERS):
    """ Constructor. """
    self.trans_dist_computer = trans_dist_computer
    self.num_dists = len(betas)
    self.num_dist_types = len(self.trans_dist_computer.dflt_dist_type.split('-'))
    # Call super constructor
    super(DistProdNNKernel, self).__init__(self.compute_dists, betas, scale,
                                           powers, self.num_dists)

  def compute_dists(self, X1, X2):
    """ Given two lists of neural networks computes the distance between the two. """
    return self.trans_dist_computer(X1, X2)

  def _child_gradient(self, param, X1, X2, param_num=None):
    """ Computes gradient of kernel w.r.t the param """
    # TODO: Need to implement this.
    raise NotImplementedError('Should implement this gradient!')


class DistSumNNKernel(SumOfExpSumOfDistsKernel):
  """ Computes a kernel using the transportation distance as the distance. """

  def __init__(self, trans_dist_computer, alphas, betas, powers=DFLT_KERN_DIST_POWERS):
    """ Constructor. """
    self.trans_dist_computer = trans_dist_computer
    self.num_dists = len(betas)
    self.num_dist_types = len(self.trans_dist_computer.dflt_dist_type.split('-'))
    groups = self._get_groups(self.num_dists, self.num_dist_types)
    # Call super constructor
    super(DistSumNNKernel, self).__init__(self.compute_dists, alphas, groups, betas,
                                          powers)

  @classmethod
  def _get_groups(cls, num_dists, num_dist_types):
    """ Returns the number of groups. """
    num_dists_per_group = num_dists//num_dist_types
    groups = [[(i + num_dist_types * j) for j in range(num_dists_per_group)]
              for i in range(num_dist_types)]
    return groups

  def compute_dists(self, X1, X2):
    """ Given two lists of neural networks computes the distance between the two. """
    return self.trans_dist_computer(X1, X2)

  def _child_gradient(self, param, X1, X2, param_num=None):
    """ Computes gradient of kernel w.r.t the param """
    # TODO: Need to implement this.
    raise NotImplementedError('Should implement this gradient!')


# APIs to return a distance computer or kernel -------------------------------------------
def get_otmann_distance_computer_from_args(nn_type, non_assignment_penalty,
      mislabel_coeffs, struct_coeffs, dist_type,
      connectivity_diff_cost_function=DFLT_CONN_COST_FUNC):
  """ Obtains a transport distance computer from dists. """
  if nn_type.startswith('cnn'):
    all_layer_labels, label_mismatch_penalty = \
      get_cnn_layer_label_mismatch_penalties(non_assignment_penalty)
    struct_penalty_groups = CNN_STRUCTURAL_PENALTY_GROUPS
  elif nn_type.startswith('mlp'):
    all_layer_labels, label_mismatch_penalty = \
      get_mlp_layer_label_mismatch_penalties(non_assignment_penalty,
                    nn_type[4:])
    struct_penalty_groups = MLP_STRUCTURAL_PENALTY_GROUPS
  # Now create a tp_comp object
  tp_comp = OTMANNDistanceComputer(all_layer_labels,
              label_mismatch_penalty, non_assignment_penalty,
              struct_penalty_groups, PATH_LENGTH_TYPES,
              dflt_mislabel_coeffs=mislabel_coeffs,
              dflt_struct_coeffs=struct_coeffs,
              dflt_dist_type=dist_type,
              connectivity_diff_cost_function=connectivity_diff_cost_function)
  return tp_comp


def get_default_otmann_distance_computer(nn_type, non_assignment_penalty=1.0,
                                         mislabel_coeffs='default',
                                         struct_coeffs='default',
                                         dist_type='lp-emd'):
  """ The otmann distance with default parameters. """
  if mislabel_coeffs == 'default':
    mislabel_coeffs = [1.0, 1.0, 1.0, 1.0]
  if struct_coeffs == 'default':
    struct_coeffs = [0.1, 0.25, 0.61, 1.5]
  return get_otmann_distance_computer_from_args(nn_type, non_assignment_penalty,
           mislabel_coeffs, struct_coeffs, dist_type)


def get_otmann_kernel_from_params(
  kernel_type, # The kernel type, should be 'sum' or 'product'
  all_layer_labels, label_mismatch_penalty, # mandatory distance args
  non_assignment_penalty, structural_penalty_groups, path_length_types,
  mislabel_coeffs, struct_coeffs, dist_type, betas, scales, # mandatory kernel args
  connectivity_diff_cost_function=DFLT_CONN_COST_FUNC, # optional dist args
  powers=DFLT_KERN_DIST_POWERS): # optional kernel args
  #pylint: disable=too-many-arguments
  """ Generates a OTMANNKernel from all parameters for the distance computer
      and the kernel.
      scales should be a scalar if kernel_type is 'prod' and a list if 'sum'.
  """
  tp_comp = OTMANNDistanceComputer(all_layer_labels, label_mismatch_penalty,
              non_assignment_penalty, structural_penalty_groups, path_length_types,
              dflt_mislabel_coeffs=mislabel_coeffs,
              dflt_struct_coeffs=struct_coeffs,
              dflt_dist_type=dist_type,
              connectivity_diff_cost_function=connectivity_diff_cost_function)
  if kernel_type == 'prod':
    return DistProdNNKernel(tp_comp, betas, scales, powers)
  elif kernel_type == 'sum':
    return DistSumNNKernel(tp_comp, scales, betas, powers)




"""
  Implements some instances of a random optimiser.
  -- kandasamy@cs.cmu.edu
"""

# pylint: disable=invalid-name

from argparse import Namespace
import numpy as np
# Local imports
from ..exd.cp_domain_utils import get_processed_func_from_raw_func_for_cp_domain, \
                                load_cp_domain_from_config_file
from ..exd import domains
from ..exd.exd_utils import get_euclidean_initial_qinfos, get_cp_domain_initial_qinfos
from ..exd.exd_core import mf_exd_args
from ..exd.experiment_caller import CPFunctionCaller
from ..exd.cp_domain_utils import sample_from_cp_domain
from .blackbox_optimiser import BlackboxOptimiser, blackbox_opt_args, \
                                   CalledMFOptimiserWithSFCaller
from ..utils.option_handler import load_options
from ..utils.reporters import get_reporter
from ..utils.general_utils import map_to_bounds

random_optimiser_args = blackbox_opt_args
euclidean_random_optimiser_args = random_optimiser_args
cp_random_optimiser_args = random_optimiser_args
mf_euclidean_random_optimiser_args = euclidean_random_optimiser_args + mf_exd_args
mf_cp_random_optimiser_args = cp_random_optimiser_args + mf_exd_args


# Base class for Random Optimisation -----------------------------------------------
class RandomOptimiser(BlackboxOptimiser):
  """ A class which optimises using random evaluations. """
  #pylint: disable=attribute-defined-outside-init
  #pylint: disable=abstract-method

  # Constructor.
  def __init__(self, func_caller, worker_manager, options=None, reporter=None):
    """ Constructor. """
    reporter = get_reporter(reporter)
    if options is None:
      options = load_options(random_optimiser_args, reporter=reporter)
    super(RandomOptimiser, self).__init__(func_caller, worker_manager, model=None,
                                          options=options, reporter=reporter)

  def _opt_method_set_up(self):
    """ Any set up specific to otptimisation. """
    pass

  def _get_method_str(self):
    """ Returns a string describing the method. """
    return 'rand'

  def _add_data_to_model(self, qinfos):
    """ Adds data to model. """
    pass

  def _child_build_new_model(self):
    """ Builds a new model. """
    pass


# Random optimiser for Euclidean spaces --------------------------------------------
class EuclideanRandomOptimiser(RandomOptimiser):
  """ A class which optimises in Euclidean spaces using random evaluations. """

  def is_an_mf_method(self):
    """ Returns False since this is not a MF method. """
    return False

  def _determine_next_query(self):
    """ Determines the next query. """
    qinfo = Namespace(point=map_to_bounds(np.random.random(self.domain.dim),
                                          self.domain.bounds))
    return qinfo

  def _determine_next_batch_of_queries(self, batch_size):
    """ Determines the next batch of queries. """
    qinfos = [self._determine_next_query() for _ in range(batch_size)]
    return qinfos

  def _get_initial_qinfos(self, num_init_evals):
    """ Returns initial qinfos. """
    return get_euclidean_initial_qinfos(self.options.init_method, num_init_evals,
                                        self.domain.bounds)


# Multi-fidelity Random Optimiser for Euclidean Spaces -------------------------------
class MFEuclideanRandomOptimiser(RandomOptimiser):
  """ A class which optimises in Euclidean spaces using random evaluations and
      multi-fidelity.
  """

  # Constructor.
  def __init__(self, func_caller, worker_manager, call_fidel_to_opt_prob=0.25,
               *args, **kwargs):
    """ Constructor.
        call_fidel_to_opt_prob is the probability with which we will choose
        fidel_to_opt as the fidel.
    """
    super(MFEuclideanRandomOptimiser, self).__init__(func_caller, worker_manager,
                                                     *args, **kwargs)
    self.call_fidel_to_opt_prob = call_fidel_to_opt_prob
    if not func_caller.is_mf():
      raise CalledMFOptimiserWithSFCaller(self, func_caller)

  def is_an_mf_method(self):
    """ Returns Truee since this is a MF method. """
    return True

  def _determine_next_query(self):
    """ Determines the next query. """
    # An internal function which returns the next fidelity.
    def _get_next_fidel():
      """ Returns the next fidelity. """
      if np.random.random() <= self.call_fidel_to_opt_prob:
        return self.func_caller.fidel_to_opt
      else:
        return np.random.random(self.fidel_space.dim)
    # Create and return qinfo
    qinfo = Namespace(point=np.random.random(self.domain.dim),
                      fidel=_get_next_fidel())
    return qinfo

  def _determine_next_batch_of_queries(self, batch_size):
    """ Determines the next batch of queries. """
    qinfos = [self._determine_next_query() for _ in range(batch_size)]
    return qinfos

  def _get_initial_qinfos(self, num_init_evals):
    """ Returns initial qinfos. """
    return get_euclidean_initial_qinfos(self.options.init_method, num_init_evals,
             self.domain.bounds, self.options.fidel_init_method, self.fidel_space.bounds,
             self.func_caller.fidel_to_opt,
             self.options.init_set_to_fidel_to_opt_with_prob)


# A random optimiser in Cartesian product spaces --------------------------------------
class CPRandomOptimiser(RandomOptimiser):
  """ A random optimiser for cartesian product domains. """

  def is_an_mf_method(self):
    """ Returns False since it is not a False method. """
    return False

  def _determine_next_query(self):
    """ Determines the next query. """
    qinfo = Namespace(point=sample_from_cp_domain(self.domain, 1)[0])
    return qinfo

  def _determine_next_batch_of_queries(self, batch_size):
    """ Determines the next batch of queries. """
    qinfos = [self._determine_next_query() for _ in range(batch_size)]
    return qinfos

  def _get_initial_qinfos(self, num_init_evals):
    """ Returns initial qinfos. """
    return get_cp_domain_initial_qinfos(self.domain, num_init_evals,
                                        dom_euclidean_sample_type='latin_hc',
                                        dom_integral_sample_type='latin_hc',
                                        dom_nn_sample_type='rand')


class MFCPRandomOptimiser(RandomOptimiser):
  """ A class which optimises in Cartesian product spaces using random evaluations. """

  def __init__(self, func_caller, worker_manager, call_fidel_to_opt_prob=0.25,
               *args, **kwargs):
    """ Constructor. """
    super(MFCPRandomOptimiser, self).__init__(func_caller, worker_manager,
                                              *args, **kwargs)
    self.call_fidel_to_opt_prob = call_fidel_to_opt_prob
    if not func_caller.is_mf():
      raise CalledMFOptimiserWithSFCaller(self, func_caller)

  def is_an_mf_method(self):
    """ Returns False since it is not a False method. """
    return True

  def _determine_next_query(self):
    """ Determines next query. """
    # An internal function which returns the next fidelity.
    def _get_next_fidel():
      """ Returns the next fidelity. """
      if np.random.random() <= self.call_fidel_to_opt_prob:
        return self.func_caller.fidel_to_opt
      else:
        return sample_from_cp_domain(self.func_caller.fidel_space, 1)[0]
    # Create and return qinfo
    qinfo = Namespace(point=sample_from_cp_domain(self.func_caller.domain, 1)[0],
                      fidel=_get_next_fidel())
    return qinfo

  def _determine_next_batch_of_queries(self, batch_size):
    """ Determines the next batch of queries. """
    qinfos = [self._determine_next_query() for _ in range(batch_size)]
    return qinfos

  def _get_initial_qinfos(self, num_init_evals):
    """ Returns initial qinfos. """
    return get_cp_domain_initial_qinfos(self.domain, num_init_evals,
             self.fidel_space, self.func_caller.fidel_to_opt,
             set_to_fidel_to_opt_with_prob=self.call_fidel_to_opt_prob,
             dom_euclidean_sample_type='latin_hc',
             dom_integral_sample_type='latin_hc',
             dom_nn_sample_type='rand',
             fidel_space_euclidean_sample_type='latin_hc',
             fidel_space_integral_sample_type='latin_hc',
             fidel_space_nn_sample_type='rand',
             )


# APIs for random optimisation ===========================================================

# An API for single fidelity optimisation
def random_optimiser_from_func_caller(func_caller, worker_manager, max_capital,
                                      mode='asy', options=None, reporter='default'):
  """ Creates an appropriate RandomOptimiser Object and optimises the function. """
  reporter = get_reporter(reporter)
  if isinstance(func_caller.domain, domains.EuclideanDomain):
    optimiser_constructor = EuclideanRandomOptimiser
    dflt_list_of_options = euclidean_random_optimiser_args
  elif isinstance(func_caller.domain, domains.CartesianProductDomain):
    optimiser_constructor = CPRandomOptimiser
    dflt_list_of_options = cp_random_optimiser_args
  else:
    raise ValueError('Random optimiser not implemented for domain of type %s.'%(
                     type(func_caller.domain)))
  # Load options and modify where necessary
  if options is None:
    options = load_options(dflt_list_of_options)
  options.mode = mode
  from ..exd.worker_manager import RealWorkerManager, SyntheticWorkerManager
  if isinstance(worker_manager, RealWorkerManager):
    options.capital_type = 'realtime'
  elif isinstance(worker_manager, SyntheticWorkerManager):
    options.capital_type = 'return_value'
  # Create optimiser
  optimiser = optimiser_constructor(func_caller, worker_manager, options, reporter)
  # optimise and return
  return optimiser.optimise(max_capital)

def cp_random_optimiser_from_raw_args(raw_func, domain_config_file, *args, **kwargs):
  """ A random optimiser on Cartesian product spaces. """
  # pylint: disable=no-member
  cp_dom, orderings = load_cp_domain_from_config_file(domain_config_file)
  proc_func = get_processed_func_from_raw_func_for_cp_domain(
                raw_func, cp_dom, orderings.index_ordering, orderings.dim_ordering)
  func_caller = CPFunctionCaller(proc_func, cp_dom, raw_func=raw_func,
                                 domain_orderings=orderings)
  return random_optimiser_from_func_caller(func_caller, *args, **kwargs)


# An API for multi-fidelity optimisation
def mf_random_optimiser_from_func_caller(func_caller, worker_manager, max_capital,
                                         mode='asy', options=None, reporter='default',
                                         *args, **kwargs):
  """ Creates a MF EuclideanRandomOptimiser Object and optimises the function. """
  reporter = get_reporter(reporter)
  if isinstance(func_caller.domain, domains.EuclideanDomain) and \
     isinstance(func_caller.fidel_space, domains.EuclideanDomain):
    optimiser_constructor = MFEuclideanRandomOptimiser
    dflt_list_of_options = mf_euclidean_random_optimiser_args
  elif isinstance(func_caller.domain, domains.CartesianProductDomain) and \
    isinstance(func_caller.fidel_space, domains.CartesianProductDomain):
    optimiser_constructor = MFCPRandomOptimiser
    dflt_list_of_options = mf_cp_random_optimiser_args
  else:
    raise ValueError(('MF Random optimiser not implemented for (domain, fidel_space) '
                      + 'of types (%s, %s).')%(
                     type(func_caller.domain), type(func_caller.fidel_space)))
  # Load options
  if options is None:
    options = load_options(dflt_list_of_options)
  options.mode = mode
  # Create optimiser
  optimiser = optimiser_constructor(func_caller, worker_manager, options=options,
                                    reporter=reporter, *args, **kwargs)
  # optimise and return
  return optimiser.optimise(max_capital)

